#! /usr/bin/env python
# encoding: utf-8
"""
# test_load_numpy

Unit tests for hickle module -- numpy loader.

"""
import pytest

import sys

# %% IMPORTS
# Package imports
import h5py as h5
import numpy as np
import hickle.loaders.load_numpy as load_numpy
from py.path import local


# Set current working directory to the temporary directory
local.get_temproot().chdir()

# %% GLOBALS

NESTED_DICT = {
    "level1_1": {
        "level2_1": [1, 2, 3],
        "level2_2": [4, 5, 6]
    },
    "level1_2": {
        "level2_1": [1, 2, 3],
        "level2_2": [4, 5, 6]
    },
    "level1_3": {
        "level2_1": {
            "level3_1": [1, 2, 3],
            "level3_2": [4, 5, 6]
        },
        "level2_2": [4, 5, 6]
    }
}

# %% FIXTURES

@pytest.fixture
def h5_data(request):
    """
    create dummy hdf5 test data file for testing PyContainer and H5NodeFilterProxy
    """
    dummy_file = h5.File('test_load_builtins.hdf5','w')
    dummy_file = h5.File('load_numpy_{}.hdf5'.format(request.function.__name__),'w')
    filename = dummy_file.filename
    test_data = dummy_file.create_group("root_group")
    yield test_data
    dummy_file.close()

# %% FUNCTION DEFINITIONS

def test_create_np_scalar(h5_data,compression_kwargs):
    """
    tests proper storage and loading of numpy scalars
    """

    # check that scalar dataset is created for numpy scalar
    scalar_data = np.float64(np.pi)
    dtype = scalar_data.dtype
    h_dataset,subitems = load_numpy.create_np_scalar_dataset(scalar_data,h5_data,"scalar_data",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and iter(subitems) and not subitems
    assert h_dataset.attrs['np_dtype'] in ( dtype.str.encode('ascii'),dtype.str)
    assert  h_dataset[()] == scalar_data
    assert load_numpy.load_np_scalar_dataset(h_dataset,b'np_scalar',scalar_data.__class__) == scalar_data

    # check that numpy.bool_ scarlar is properly stored and reloaded
    scalar_data = np.bool_(True)
    dtype = scalar_data.dtype
    h_dataset,subitems = load_numpy.create_np_scalar_dataset(scalar_data,h5_data,"generic_data",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and iter(subitems) and not subitems
    assert h_dataset.attrs['np_dtype'] in ( dtype.str.encode('ascii'),dtype.str) and h_dataset[()] == scalar_data
    assert load_numpy.load_np_scalar_dataset(h_dataset,b'np_scalar',scalar_data.__class__) == scalar_data

def test_create_np_dtype(h5_data,compression_kwargs):
    """
    test proper creation and loading of dataset representing numpy dtype
    """ 
    dtype = np.dtype(np.int16)
    h_dataset,subitems = load_numpy.create_np_dtype(dtype, h5_data,"dtype_string",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and iter(subitems) and not subitems
    assert bytes(h_dataset[()]).decode('ascii') == dtype.str
    assert load_numpy.load_np_dtype_dataset(h_dataset,'np_dtype',np.dtype) == dtype

def test_create_np_ndarray(h5_data,compression_kwargs):
    """
    test proper creation and loading of numpy ndarray
    """

    # check that numpy array representing python utf8 string is properly 
    # stored as bytearray dataset and reloaded from
    np_array_data = np.array("im python string")
    h_dataset,subitems = load_numpy.create_np_array_dataset(np_array_data,h5_data,"numpy_string_array",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and iter(subitems) and not subitems
    assert bytes(h_dataset[()]) == np_array_data.tolist().encode("utf8")
    assert h_dataset.attrs["np_dtype"] in ( np_array_data.dtype.str.encode("ascii"),np_array_data.dtype.str)
    assert load_numpy.load_ndarray_dataset(h_dataset,b'ndarray',np.ndarray) == np_array_data

    # check that numpy array representing python bytes string is properly
    # stored as bytearray dataset and reloaded from
    np_array_data = np.array(b"im python bytes")
    h_dataset,subitems = load_numpy.create_np_array_dataset(np_array_data,h5_data,"numpy_bytes_array",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and iter(subitems) and not subitems
    assert h_dataset[()] == np_array_data.tolist()
    assert h_dataset.attrs["np_dtype"] in ( np_array_data.dtype.str.encode("ascii"),np_array_data.dtype.str)
    assert load_numpy.load_ndarray_dataset(h_dataset,b'ndarray',np.ndarray) == np_array_data

    # check that numpy array with dtype object representing list of various kinds
    # of objects is converted to list before storing and reloaded properly from this
    # list representation

    # NOTE: simplified as mixing items of varying length receives
    # VisibleDeprecationWarning from newer numpy versions
    #np_array_data = np.array([[NESTED_DICT], ('What is this?',), {1, 2, 3, 7, 1}])
    np_array_data = np.array([NESTED_DICT])#, ('What is this?',), {1, 2, 3, 7, 1}])
    h_dataset,subitems = load_numpy.create_np_array_dataset(np_array_data,h5_data,"numpy_list_object_array",**compression_kwargs)
    ndarray_container = load_numpy.NDArrayLikeContainer(h_dataset.attrs,b'ndarray',np_array_data.__class__)
    assert isinstance(h_dataset,h5.Group) and iter(subitems)
    assert h_dataset.attrs["np_dtype"] in ( np_array_data.dtype.str.encode("ascii"),np_array_data.dtype.str)
    for index,(name,item,attrs,kwargs) in enumerate(subitems):
        assert name == "data{:d}".format(index) and attrs.get("item_index",None) == index
        assert isinstance(kwargs,dict) and np_array_data[index] == item
        ndarray_container.append(name,item,attrs)
    assert np.all(ndarray_container.convert() == np_array_data)

    # check that numpy array containing multiple strings of length > 1
    # is properly converted to list of strings and restored from its list
    # representation
    np_array_data = np.array(["1313e", "was", "maybe?", "here"])
    h_dataset,subitems = load_numpy.create_np_array_dataset(np_array_data,h5_data,"numpy_list_of_strings_array",**compression_kwargs)
    ndarray_container = load_numpy.NDArrayLikeContainer(h_dataset.attrs,b'ndarray',np_array_data.__class__)
    assert isinstance(h_dataset,h5.Group) and iter(subitems)
    assert h_dataset.attrs["np_dtype"] in ( np_array_data.dtype.str.encode("ascii"),np_array_data.dtype.str)
    for index,(name,item,attrs,kwargs) in enumerate(subitems):
        assert name == "data{:d}".format(index) and attrs.get("item_index",None) == index
        assert isinstance(kwargs,dict) and np_array_data[index] == item
        ndarray_container.append(name,item,attrs)
    assert np.all(ndarray_container.convert() == np_array_data)

    # check that numpy array with object dtype which is converted to single object
    # by ndarray.tolist method is properly stored according to type of object and
    # restored from this representation accordingly
    np_array_data = np.array(NESTED_DICT)
    h_dataset,subitems = load_numpy.create_np_array_dataset(np_array_data,h5_data,"numpy_object_array",**compression_kwargs)
    ndarray_container = load_numpy.NDArrayLikeContainer(h_dataset.attrs,b'ndarray',np_array_data.__class__)
    ndarray_pickle_container = load_numpy.NDArrayLikeContainer(h_dataset.attrs,b'ndarray',np_array_data.__class__)
    assert isinstance(h_dataset,h5.Group) and iter(subitems)
    assert h_dataset.attrs["np_dtype"] in ( np_array_data.dtype.str.encode("ascii"),np_array_data.dtype.str)
    data_set = False
    for name,item,attrs,kwargs in subitems:
        if name == "data":
            assert not data_set and not attrs and isinstance(kwargs,dict)
            assert np_array_data[()] == item
            data_set = True
            ndarray_container.append(name,item,attrs)
            attrs = dict(attrs)
            attrs["base_type"] = b'pickle'
            ndarray_pickle_container.append(name,item,attrs)
        else:
            raise AssertionError("expected single data object")
    assert np.all(ndarray_container.convert() == np_array_data)
    assert np.all(ndarray_pickle_container.convert() == np_array_data)

    # check that numpy.matrix type object is properly stored and reloaded from
    # hickle file.
    # NOTE/TODO: current versions of numpy issue PendingDeprecationWarning when using
    # numpy.matrix. In order to indicate to pytest that this is known and can safely
    # be ignored the warning is captured here. Shall it be that future numpy versions
    # convert PendingDeprecationWarning into any kind of exception like TypeError
    # AttributeError, RuntimeError or alike that also capture these Exceptions not
    # just PendingDeprecationWarning
    with pytest.warns(PendingDeprecationWarning):
        np_array_data = np.matrix([[1, 2], [3, 4]])
        h_dataset,subitems = load_numpy.create_np_array_dataset(np_array_data,h5_data,"numpy_matrix",**compression_kwargs)
        assert isinstance(h_dataset,h5.Dataset) and iter(subitems) and not subitems
        assert np.all(h_dataset[()] == np_array_data)
        assert h_dataset.attrs["np_dtype"] in ( np_array_data.dtype.str.encode("ascii"),np_array_data.dtype.str)
        np_loaded_array_data = load_numpy.load_ndarray_dataset(h_dataset,b'npmatrix',np.matrix)
        assert np.all(np_loaded_array_data == np_array_data)
        assert isinstance(np_loaded_array_data,np.matrix)
        assert np_loaded_array_data.shape == np_array_data.shape
    
def test_create_np_masked_array(h5_data,compression_kwargs):
    """
    test proper creation and loading of numpy.masked arrays
    """

    # check that simple masked array is properly stored and loaded
    masked_array = np.ma.array([1, 2, 3, 4], dtype='float32', mask=[0, 1, 0, 0])
    h_datagroup,subitems = load_numpy.create_np_masked_array_dataset(masked_array, h5_data, "masked_array",**compression_kwargs)
    masked_array_container = load_numpy.NDMaskedArrayContainer(h_datagroup.attrs,b'ndarray_masked',np.ma.array)
    assert isinstance(h_datagroup,h5.Group) and iter(subitems)
    assert h_datagroup.attrs["np_dtype"] in ( masked_array.dtype.str.encode("ascii"),masked_array.dtype.str)
    data_set = mask_set = False
    for name,item,attrs,kwargs in subitems:
        assert isinstance(attrs,dict) and isinstance(kwargs,dict)
        if name == "data":
            assert not data_set and not attrs and np.all(masked_array.data == item) and item is not masked_array
            masked_array_container.append(name,item,attrs)
            data_set = True
        elif name == "mask":
            assert not mask_set and not attrs and np.all(masked_array.mask == item) and item is not masked_array
            masked_array_container.append(name,item,attrs)
            mask_set = True
        else:
            raise AssertionError("expected one data and one mask object")
    assert np.all(masked_array_container.convert() == masked_array)

    # check that format used by hickle version 4.0.0 to encode is properly recognized
    # on loading and masked array is restored accordingly
    h_dataset = h5_data.create_dataset("masked_array_dataset",data = masked_array.data)
    h_dataset.attrs["np_dtype"] = masked_array.dtype.str.encode("ascii")
    with pytest.raises(ValueError,match = r"mask\s+not\s+found"):
        loaded_masked_array = load_numpy.load_ndarray_masked_dataset(h_dataset,b'masked_array_data',np.ma.array)
    h_mask_dataset = h5_data.create_dataset("masked_array_dataset_mask",data = masked_array.mask)
    loaded_masked_array = load_numpy.load_ndarray_masked_dataset(h_dataset,b'masked_array_data',np.ma.array)
    assert np.all(loaded_masked_array == masked_array )
    
# %% MAIN SCRIPT
if __name__ == "__main__":
    from _pytest.fixtures import FixtureRequest
    from hickle.tests.conftest import compression_kwargs
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_create_np_scalar),)
    ):
        test_create_np_scalar(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_create_np_dtype),)
    ):
        test_create_np_dtype(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_create_np_ndarray),)
    ):
        test_create_np_ndarray(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_create_np_masked_array),)
    ):
        test_create_np_masked_array(h5_root,keywords)

    



# hickle imports
from .__version__ import __version__
from . import hickle
from .hickle import *
from .fileio import ClosedFileError, FileError

# All declaration
__all__ = ['hickle', 'ClosedFileError', 'FileError']
__all__.extend(hickle.__all__)

# Author declaration
__author__ = "Danny Price, Ellert van der Velden and contributors"



import pytest
import sys
import types
import functools as ft
import threading
import os
import os.path
import importlib
import collections
import ctypes
import re

# list of function names which shall not be
# traced when compression keyword hardening
# test run is executed
non_core_loader_functions = {
    'type_legacy_mro',
    'load_pickled_data',
    'recover_custom_dataset',
    #'register_compact_expand',
    '_moc_numpy_array_object_lambda',
    'fix_lambda_obj_type',
    'LoaderManager.load_loader',
    'RecoverGroupContainer.convert',
    'NoContainer.convert',
    '_DictItemContainer.convert',
    'ExpandReferenceContainer.convert',
    'RecoverGroupContainer.filter',
    'ExpandReferenceContainer.filter',
    'ReferenceManager.resolve_type',
    'RecoverGroupContainer._append'
}

def pytest_addoption(parser):
    """
    adds enable_compression keywort to pytest commandline options
    for enabling h5py compression keyword hardening testing of
    dump functions of hikcle.loaders  and hickle core loaders
    """
    parser.addoption(
        "--enable-compression",
        action='store',
        nargs='?',
        const=6,
        type=int,
        choices=range(0,10),
        help="run all tests with bzip compression enabled. Optionally specify compression level 0-9 (default 6)",
        dest="enable_compression"
    )

def _get_trace_function(trace_function):
    """
    try to get hold of FunctionType object of passed in Method, Function or callable
    """
    while not isinstance(trace_function,(types.FunctionType,types.LambdaType,types.BuiltinFunctionType)):
        if isinstance(trace_function,(types.MethodType,types.BuiltinMethodType)):
            trace_function = getattr(trace_function,'__func__')
            continue
        if isinstance(trace_function,ft.partial):
            trace_function = trace_function.func
            continue
        return (
            getattr(trace_function,'__call__',trace_function) 
            if callable(trace_function) and not isinstance(trace_function,type) else
            None
        )
    return trace_function

# keyword arguments to yield from compression_kwargs fixture below
# may in future become a list of dictionaries to be yieled for
# running same test with different sets of compression keywords
# (implizit parametrization of tests)
_compression_args =  dict(
    compression='gzip',
    compression_opts=6
)
_test_compression = None

def pytest_configure(config):
    """
    make no_compression mark available from pytest.mark.
    if not yet activated enable profiling of dump methods and functions
    and set compression_level selected on commandline if explicitly
    specified.
    """
    global _test_compression

    config.addinivalue_line(
        "markers","no_compression: do not enforce h5py compression hardening testing"
    )
    if _test_compression is not None:
        return
    compression_level = config.getoption("enable_compression",default=-1)
    if compression_level is None or compression_level < 0:
        return
    _compression_args['compression_opts'] = compression_level
    _test_compression = True

# local handle of no_compression mark
no_compression = pytest.mark.no_compression

@pytest.fixture#(scope='session')
def compression_kwargs(request):
    """
    fixture providing the compressoin related keyword arguments
    to be passed to any test not marked with no_compression mark
    and expecting compression_kwargs as one of its parameters 
    """
    global _test_compression
    yield ( _compression_args if _test_compression else {} )

# list of distinct copyies of LoaderManager.register_class function
# keys are either "<filename>::LoaderManager.register_class" or
# copy of code object executed when LoaderManager.register_class method
# is called
_trace_register_class = {}

# list of dump_functions to be traced with respect to being
# passed the compression related keywords provided through compression_kwargs
# fixture above. In case a call to any of these does not include at least these
# keywords an AssertionError Exception is raised.
_trace_functions = collections.OrderedDict()

# profiling function to be called after execution of _trace_loader_funcs
# below
_trace_profile_call = None

# index of dump_function argument in argument list of LoaderManager.register_class
# method. 
_trace_function_argument_default = -1
def _chain_profile_call(frame,event,arg):
    global _trace_profile_call
    if _trace_profile_call:
        next_call = _trace_profile_call(frame,event,arg)
        if next_call:
            _trace_profile_call = next_call

# argument names which correspond to argument being passed dump_function
# object
_trace_function_arg_names = {'dump_function'}


# the pytest session tracing of proper handling of compression related
# keywords is activated for
_traced_session = None

_loader_file_pattern = re.compile(r'^load_\w+\.py$')

def pytest_sessionstart(session):
    """
    pytest hook called at start of session.
    - collects all functions exported by hickle.lookup module (for now) and
      records inserts "<filename>::<function.__qualname__>" strings into
      _trace_functions list for any not listed in above non_core_loader_functions
    - collects all dump_functions listed in class_register tables of all
      hickle.loaders.load_*.py modules.
    """
    global _test_compression,_traced_session,_trace_register_class,_trace_functions,_trace_profile_call
    if _test_compression is None:
        pytest_configure(session.config)
    if not _test_compression:
        return None
    # extract all loader function from hickle.lookup
    lookup_module = sys.modules.get('hickle.lookup',None)
    if not isinstance(lookup_module,types.ModuleType):
        lookup_module_spec = importlib.util.find_spec("hickle.lookup")
        
        lookup_module = importlib.util.module_from_spec(lookup_module_spec)
        lookup_module_spec.loader.exec_module(lookup_module)
    register_class = lookup_module.LoaderManager.register_class
    register_class_code = register_class.__func__.__code__
    trace_function_argument = register_class_code.co_argcount + register_class_code.co_kwonlyargcount
    for argid,trace_function in ( (count,varname) for count,varname in enumerate(register_class_code.co_varnames[:(register_class_code.co_argcount + register_class_code.co_kwonlyargcount)]) if varname in _trace_function_arg_names ):
        trace_function_argument = argid
        break
    if trace_function_argument < 0:
        return None
    _trace_function_argument_default = trace_function_argument
    qualname = getattr(register_class,'__qualname__',register_class.__name__)
    code_name = qualname if qualname.rsplit('.',1) == register_class_code.co_name else register_class_code.co_name
    _trace_register_class.update({"{}::{}".format(register_class_code.co_filename,code_name):trace_function_argument})
    for loader_func_name,loader_func in ( 
        (func_name,func) 
        for name, item in lookup_module.__dict__.items() 
        if isinstance(item,(types.FunctionType,type))
        for func_name,func in ( 
            ((name,item),)
            if isinstance(item,types.FunctionType) else 
            ( 
                ( meth_name,meth) 
                for meth_name,meth in item.__dict__.items() 
                if isinstance(meth,types.FunctionType) 
            ) 
        )
        if func_name[:2] != '__' and func_name[-2:] != '__' 
    ):
        loader_func = _get_trace_function(loader_func)
        if loader_func is not None and loader_func.__module__ == lookup_module.__name__:
            code = loader_func.__code__
            qualname = getattr(loader_func,'__qualname__',loader_func.__name__) 
            if qualname not in non_core_loader_functions:
                code_name = qualname if qualname.rsplit('.',1) == code.co_name else code.co_name
                _trace_functions["{}::{}".format(code.co_filename,code_name)] = (loader_func.__module__,qualname)
    # extract all dump functions from any known loader module
    hickle_loaders_path = os.path.join(os.path.dirname(lookup_module.__file__),'loaders')
    for loader in os.scandir(hickle_loaders_path):
        if not loader.is_file() or _loader_file_pattern.match(loader.name) is None:
            continue
        loader_module_name = "hickle.loaders.{}".format(loader.name.rsplit('.',1)[0])
        loader_module = sys.modules.get(loader_module_name,None)
        if loader_module is None:
            
            loader_module_spec = importlib.util.find_spec("hickle.loaders.{}".format(loader.name.rsplit('.',1)[0]))
            if loader_module_spec is None:
                continue
            loader_module = importlib.util.module_from_spec(loader_module_spec)
            try:
                loader_module_spec.loader.exec_module(loader_module)
            except ModuleNotFoundError:
                continue
            except ImportError:
                if sys.version_info[0] > 3 or sys.version_info[1] > 5:
                    raise
                continue
        class_register_table = getattr(loader_module,'class_register',())
        # trace function has cls/self
        for dump_function in ( entry[trace_function_argument-1] for entry in class_register_table ):
            dump_function = _get_trace_function(dump_function)
            if dump_function is not None:
                code = dump_function.__code__
                qualname = getattr(dump_function,'__qualname__',dump_function.__name__)
                code_name = qualname if qualname.rsplit('.',1) == code.co_name else code.co_name
                _trace_functions["{}::{}".format(code.co_filename,code_name)] = (dump_function.__module__,qualname)
    # activate compression related profiling
    _trace_profile_call = sys.getprofile()
    _traced_session = session
    sys.setprofile(_trace_loader_funcs)
    return None

# List of test functions which are marked by no_compression mark
_never_trace_compression = set()

def traceback_from_frame(frame,stopafter):
    """
    helper function used in Python >= 3.7 to beautify traceback
    of AssertionError exceptoin thrown by _trace_loader_funcs
    """
    tb = types.TracebackType(None,frame,frame.f_lasti,frame.f_lineno)
    while frame.f_back is not stopafter.f_back:
        frame = frame.f_back
        tb = types.TracebackType(tb,frame,frame.f_lasti,frame.f_lineno)
    return tb

    
def pytest_collection_finish(session):
    """
    collect all test functions for which compression related keyword monitoring
    shall be disabled.
    """
    if not sys.getprofile() == _trace_loader_funcs:
        return 

    listed = set()
    listemodules = set()
    for item in session.items:
        func = item.getparent(pytest.Function)
        if func not in listed:
            listed.add(func)
            for marker in func.iter_markers(no_compression.name):
                never_trace_code = func.function.__code__
                qualname = getattr(func.function,'__qualname__',func.function.__name__)
                code_name = qualname if qualname.rsplit('.',1) == never_trace_code.co_name else never_trace_code.co_name
                _never_trace_compression.add("{}::{}".format(never_trace_code.co_filename,code_name))
                break
        

def _trace_loader_funcs(frame,event,arg,nochain=False):
    """
    does the actuatual profiling with respect to proper passing compression keywords
    to dump_functions
    """
    global _chain_profile_call, _trace_functions,_never_trace_compression,_trace_register_class,_trace_function_argument_default
    try:
        if event not in {'call','c_call'}:
            return _trace_loader_funcs
        # check if LoaderManager.register_class has been called
        # if get position of dump_function argument and extract
        # code object for dump_function to be registered if not None
        code_block = frame.f_code
        trace_function_argument = _trace_register_class.get(code_block,None)
        if trace_function_argument is not None:
            trace_function = frame.f_locals.get(code_block.co_varnames[trace_function_argument],None)
            load_function = frame.f_locals.get(code_block.co_varnames[trace_function_argument+1],None)
            if load_function is not None:
                load_function = _get_trace_function(load_function)
                _trace_functions.pop("{}::{}".format(load_function.__code__.co_filename,load_function.__code__.co_name),None)
            if trace_function is None:
                return _trace_loader_funcs
            trace_function = _get_trace_function(trace_function)
            if trace_function is None:
                return _trace_loader_funcs
            trace_function_code = getattr(trace_function,'__code__',None)
            if trace_function_code is not None:
                # store code object corresponding to dump_function in _trace_functions list
                # if not yet present there. 
                qualname = getattr(trace_function,'__qualname__',trace_function.__name__)
                code_name = qualname if qualname.rsplit('.',1) == trace_function_code.co_name else trace_function_code.co_name
                trace_function_code_name = "{}::{}".format(trace_function_code.co_filename,code_name)
                if (
                    trace_function_code_name not in _trace_register_class and
                    ( 
                        trace_function_code_name not in _trace_functions or
                        trace_function_code not in _trace_functions
                    )
                ):
                    trace_function_spec = (trace_function.__module__,qualname)
                    _trace_functions[trace_function_code] = trace_function_spec
                    _trace_functions[trace_function_code_name] = trace_function_spec
            return _trace_loader_funcs
        # estimate qualname from local variable stored in frame.f_local corresponding
        # to frame.f_code.co_varnames[0] if any.
        object_self_name = frame.f_code.co_varnames[:1]
        if object_self_name:
            self = frame.f_locals.get(object_self_name[0],None)
            module = getattr(self,'__module__','')
            if isinstance(module,str) and module.split('.',1)[0] == 'hickle' and isinstance(getattr(self,'__name__',None),str):
                method = getattr(self,frame.f_code.co_name,None)
                if method is not None and getattr(method,'__code__',None) == frame.f_code:
                    code_name = "{}::{}.{}".format(
                        frame.f_code.co_filename,
                        getattr(self,'__qualname__',self.__name__),
                        frame.f_code.co_name
                    )
                else:
                    code_name = "{}::{}".format(frame.f_code.co_filename,frame.f_code.co_name)
            else:
                code_name = "{}::{}".format(frame.f_code.co_filename,frame.f_code.co_name)
        else:
            code_name = "{}::{}".format(frame.f_code.co_filename,frame.f_code.co_name)
        # check if frame could encode a clall to a new incarnation of LoaderManager.register_class
        # method. Add its code object to the list of known incarnations and rerun above code
        if code_block.co_name == 'register_class':
            trace_function_argument = _trace_register_class.get(code_name,None)
            if trace_function_argument is not None:
                _trace_register_class[code_block] = trace_function_argument
                return _trace_loader_funcs(frame,event,arg,True)
            if (
                code_block.co_filename.rsplit('/',2) == ['hickle','lookup.py'] and 
                code_block.co_varnames > trace_function_argument and 
                code_block.co_varnames[_trace_function_argument_default] in _trace_function_arg_names
            ):
                _trace_register_class[code_name] = _trace_function_argument_default
                _trace_register_class[code_block] = _trace_function_argument_default
                return _trace_loader_funcs(frame,event,arg,True)

        # frame encodes a call to any other function or method. 
        # If the function or method is listed in _trace_functions list check
        # if it received the appropriate set of compresson related keywords
        function_object_spec = _trace_functions.get(frame.f_code,None)
        if function_object_spec is None:
            function_object_spec = _trace_functions.get(code_name,None)
            if function_object_spec is None:
                return _trace_loader_funcs
            _trace_functions[frame.f_code] = function_object_spec
        baseargs = ( 
            (arg,frame.f_locals[arg]) 
            for arg in frame.f_code.co_varnames[:(frame.f_code.co_argcount + frame.f_code.co_kwonlyargcount)]
        )
        kwargs = frame.f_locals.get('kwargs',None)
        if kwargs is not None:
            fullargs = ( (name,arg) for arglist in (kwargs.items(),baseargs) for name,arg in arglist )
        else:
            fullargs = baseargs
        seen_compression_args = set()
        for arg,value in fullargs:
            if arg in seen_compression_args:
                continue
            if _compression_args.get(arg,None) is not None:
                seen_compression_args.add(arg)
                if len(seen_compression_args) == len(_compression_args):
                    return _trace_loader_funcs
        # keywords not passed or filtered prematurely.
        # walk the stack until reaching executed test function.
        # if test function is not marked with no_compression raise
        # AssertionError stating that dump_function did not
        # receive expected compression keywords defined above
        # For Python <= 3.6 collect all functions called between current
        # frame and frame of executed test function. For Python > 3.6 use
        # above traceback_from_frame function to build traceack showing appropriate
        # callstack and context excluding this function to ensure AssertionError
        # exception appears thrown on behlaf of function triggering call encoded by
        # passed frame
        function_object_spec = _trace_functions[frame.f_code]
        if _traced_session is not None:
            test_list = { 
                "{}::{}".format(
                    item.function.__code__.co_filename,
                    getattr(item.function,'__qualname__',
                    item.function.__name__)
                ):item 
                for item in _traced_session.items
            }
            collect_call_tree = []
            next_frame = frame
            while next_frame is not None:
                object_self_name = frame.f_code.co_varnames[:1]
                if object_self_name:
                    self = frame.f_locals.get(object_self_name[0])
                    module = getattr(self,'__module__','')
                    if (
                        isinstance(module,str) and 
                        module.split('.',1)[0] == 'hickle' and 
                        isinstance(getattr(self,'__name__',None),str)
                    ):
                        method = getattr(self,frame.f_code.co_name,None)
                        if method is not None and getattr(method,'__code__',None) == frame.f_code:
                            frame_name = "{}::{}".format(
                                next_frame.f_code.co_filename,
                                getattr(method,'__qualname__',method.__name__)
                            )
                        else:
                            frame_name = "{}::{}".format(next_frame.f_code.co_filename,next_frame.f_code.co_name)
                    else:
                        frame_name = "{}::{}".format(next_frame.f_code.co_filename,next_frame.f_code.co_name)
                else:
                    frame_name = "{}::{}".format(next_frame.f_code.co_filename,next_frame.f_code.co_name)
                if frame_name in _never_trace_compression:
                    return _trace_loader_funcs
                in_test = test_list.get(frame_name,None)
                collect_call_tree.append((next_frame.f_code.co_filename,frame_name,next_frame.f_lineno))
                if in_test is not None:
                    try:
                        tb = traceback_from_frame(frame,next_frame)
                    except TypeError:
                        pass
                    else:
                        raise AssertionError(
                            "'{}': compression_kwargs lost in call".format("::".join(function_object_spec))
                        ).with_traceback(tb)
                    raise AssertionError(
                        "'{}': compression_kwargs lost in call:\n\t{}\n".format(
                            "::".join(function_object_spec),
                            "\n\t".join("{}::{} ({})".format(*call) for call in collect_call_tree[:0:-1])
                        )
                    )
                next_frame = next_frame.f_back
    except AssertionError as ae:
        # check that first entry in traceback does not refer to this function
        if ae.__traceback__.tb_frame.f_code == _trace_loader_funcs.__code__:
            ae.__traceback__ = ae.__traceback__.tb_next
        raise
    #except Exception as e:
    #    import traceback;traceback.print_exc()
    #    import pdb;pdb.set_trace()
    finally:
        if not nochain:
            _chain_profile_call(frame,event,arg)

def pytest_sessionfinish(session):
    sys.setprofile(_trace_profile_call)



# encoding: utf-8
"""
#fileio.py

contains functions, classes and constants related to file management.
These functions may also be used by loader modules even though currently
no related use-case is known and storing dicts as independent files as
requested by @gpetty in issue #133 is better handled on hdf5 or h5py level
and not on hickle level.
"""

# %% IMPORTS
# Built-in imports
import re
import operator
import typing
import types
import collections
import numbers
import h5py as h5
import os.path as os_path
from pathlib import Path


class FileError(Exception):
    """ An exception raised if the file is fishy """


class ClosedFileError(Exception):
    """ An exception raised if the file is fishy """


def not_io_base_like(f,*args):
    """
    creates function which can be used in replacement for
    IOBase.isreadable, IOBase.isseekable and IOBase.iswriteable
    methods in case f would not provide any of them.

    Parameters
    ----------
    f (file or file like):
        file or file like object to which hickle shall
        dump data to.

    *args (tuple):
        list of one or more tuples containing the commands to be checked 
        in replacement tests for IOBase.isreadable, IOBase.isseekable or
        IOBase.iswriteable and the arguments required to perform the tests
        
        Note: **kwargs not supported

    Returns
    -------
        function to be called in replacement of any of not implemented
        IOBase.isreadable, IOBase.isseekable or IOBase.isreadable

    Example
    -------
        if not getattr(f, 'isreadable', not_io_base_like(f, 'read', 0))():
            raise ValueError("Not a readable file or file like object")
    """
    def must_test():
        if not args:
            return False
        for cmd,*call_args in args:
            cmd = getattr(f,cmd,None)
            if not cmd:
                return False
            try:
                cmd(*call_args)
            except:
                return False
        return True
    return must_test

def file_opener(f, path, mode='r',filename = None):
    """
    A file opener helper function with some error handling.
    This can open files through a file object, an h5py file, or just the
    filename.

    Parameters
    ----------
    f (file, file-like, h5py.File, str, (file,str),{'file':file,'name':str} ):
        File to open for dumping or loading purposes.
        str:
            the path of the HDF5-file that must be used.
        `~h5py.Group`:
             the group (or file) in an open HDF5-file that must be used.
        file, file-like: 
            file or like object which provides `read`, `seek`, `tell` and write methods
        tuple:
            two element tuple with the first being the file or file like object
            to dump to and the second the filename to be used instead of 'filename'
            parameter
        dict:
            dictionary with 'file' and 'name' items

    path (str):
        Path within HDF5-file or group to dump to/load from.

    mode (str): optional
        string indicating how the file shall be opened. For details see Python `open`.
        
        Note: The 'b' flag is optional as all files are a and have to be opened in
            binary mode.
    
    filename (str): optional
        The name of the file. Ignored when f is `str` or `h5py.File` object.

    Returns
    -------
    tuple containing (file, path, closeflag)

    file (h5py.File):
        The h5py.File object the data is to be dumped to or loaded from

    path (str):
        Absolute path within HDF5-file or group to dump to/load from.

    closeflag:
        True .... file was opened by file_opener and must be closed by caller.
        False ... file shall not be closed by caller unless opened by caller

    Raises
    ------
    CloseFileError:
        If passed h5py.File, h5py.Group or h5py.Dataset object is not
        accessible. This in most cases indicates that underlying HDF5.File,
        file or file-like object has already been closed.

    FileError
        If passed file or file-like object is not opened for reading or
        in addition for writing in case mode corresponds to any
        of 'w', 'w+', 'x', 'x+' or a.

    ValueError:
        If anything else than str, bytes or None specified for filename
    """

    # Make sure that the given path always starts with '/'
    if not path.startswith('/'):
        path = "/%s" % path

    # Were we handed a file object or just a file name string?
    if isinstance(f, (str, Path)):
        return h5.File(f, mode.replace('b','')),path,True
    if isinstance(f, h5.Group):
        if not f:
            raise ClosedFileError(
                "HDF5 file {}has been closed or h5py.Group or h5py.Dataset are not accessible. "
                "Please pass either a filename string, a pathlib.Path, a file or file like object, "
                "an opened h5py.File or h5py.Group or h5py.Dataset there outof.".format(
                    "'{}' ".format(filename) if isinstance(filename,(str,bytes)) and filename else ''
                )
            )
        base_path = f.name
        if not isinstance(f,h5.File):
            f = f.file
        if f.mode == 'r' and ( mode[0] != 'r' or '+' in mode[1:] ):
            raise FileError( "HDF5 file '{}' not opened for writing".format(f.filename))

        # Since this file was already open, do not close the file afterward
        return f,''.join((base_path,path.rstrip('/'))),False

    # get the name of the file
    if not isinstance(filename,(str,bytes)):
        if filename is not None:
            raise ValueError("'filename' must be of type 'str' or 'bytes'")
        if isinstance(f,(tuple,list)) and len(f) > 1:
            f,filename = f[:2]
        elif isinstance(f,dict):
            f,filename = f['file'],f['name']
        else:
            filename = getattr(f,'filename',None)
            if filename is None:
                filename = getattr(f,'name',None)
                if filename is None:
                    filename = repr(f)
    if getattr(f,'closed',False):
        raise ClosedFileError(
            "HDF5 file {}has been closed or h5py.Group or h5py.Dataset are not accessible. "
            "Please pass either a filename string, a pathlib.Path, a file or file like object, "
            "an opened h5py.File or h5py.Group or h5py.Dataset there out of.".format(
                "'{}' ".format(filename) if isinstance(filename,(str,bytes)) and filename else ''
            )
        )
    # file and file-like object must be at least read and seekable. This means
    # they have as specified by IOBase provide read, seek and tell methods
    if (
        getattr(f,'readable',not_io_base_like(f,('read',0)))() and
        getattr(f,'seekable',not_io_base_like(f,('seek',0),('tell',)))()
    ):

        # if file is to be opened for cration, writing or appending check if file, file-like
        # object is writable or at least provides write method for writing binary data
        if mode[0] in 'xwa' or ( '+' in mode[1:] and mode[0] == 'r' ):
            if not getattr(f,'writeable',not_io_base_like(f,('write',b'')))():
                raise FileError(
                    "file '{}' not writable. Please pass either a filename string, "
                    "a pathlib.Path,  a file or file like object, "
                    "an opened h5py.File or h5py.Group or h5py.Dataset there out of.".format(filename)
                )
        elif mode[0] != 'r':
            raise ValueError(
                "invalid file mode must be one out of 'w','w+','x','x+','r','r+','a'. "
                "at max including a 'b' which will be ignored"
        )
        return h5.File(
            f,
            mode.replace('b','') if mode[0] == 'r'  else mode[0],
            driver='fileobj',
            fileobj = f
        ), path, True

    raise FileError(
        "'file_obj' must be a valid path string, pahtlib.Path, h5py.File, h5py.Group, "
        "h5py.Dataset, file  or file like object'"
    )



#! /usr/bin/env python
# encoding: utf-8
"""
# test_load_pandas

Unit tests for hickle module -- pandas loader.

"""

# TODO add tests for all loader related dump_fcn, load_fcn functions
#      and PyContainer classes as soon as there exists any pandas
#      specific loader

# %% IMPORTS
# Package imports
import h5py as h5
import numpy as np
import pytest
import pandas as pd
from py.path import local

# hickle imports
import hickle.loaders.load_pandas as load_pandas


# Set the current working directory to the temporary directory
local.get_temproot().chdir()

# %% FIXTURES

@pytest.fixture
def h5_data(request):
    """
    create dummy hdf5 test data file for testing PyContainer and H5NodeFilterProxy
    """
    dummy_file = h5.File('test_load_builtins.hdf5','w')
    dummy_file = h5.File('load_numpy_{}.hdf5'.format(request.function.__name__),'w')
    filename = dummy_file.filename
    test_data = dummy_file.create_group("root_group")
    yield test_data
    dummy_file.close()

# %% FUNCTION DEFINITIONS
def test_nothing_yet_totest(h5_data,compression_kwargs):
    """
    dummy test function to be removed as soon as load_pandas loader module
    contains dump_fcn, load_fcn and PyContainer functions and classes
    for pandas arrays and objects.
    """

# %% MAIN SCRIPT
if __name__ == "__main__":
    from _pytest.fixtures import FixtureRequest
    from conftest import compression_kwargs
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_nothing_yet_totest),)
    ):
        test_nothing_yet_totest(h5_root,keywords)



# names all optional loaders defined by any load_*.py file
# will be extended by any optional loader managed by hickle
# core engine. Names of optional_loaders must be all lower case.
# Corresponding option attributes in hickle file will be all
# upper case.
optional_loaders = {
    # option loader for defining custom loader methods and 
    # PyContainer classes. By marking them as custom option
    # they are only activate if specified by a call to 
    # hickle.dump. If not specified than custom objects and
    # classes will simply be stored as pickle string.
    # The data may in this case not be recoverable if 
    # underlying classes are not available or not compatible
    # any more due to disruptive changes. When dumped using
    # custom loader hickle at least can try to restore data
    # as numpy.array or python dict like structure with metadata
    # attached as is for further inspection.
    'custom',
}

# prefix for optional_loaders attribute names which are all
# uppercase
attribute_prefix = "OPTION_"



#! /usr/bin/env python
# encoding: utf-8
"""
# test_hickle.py

Unit tests for hickle module.

"""


# %% IMPORTS
# Built-in imports
from collections import OrderedDict as odict
import os
import re
from pprint import pprint


# Package imports
import pytest
import pickle
import h5py
import numpy as np
from py.path import local

# hickle imports
from hickle import dump, helpers, hickle, load, lookup, fileio

# Set current working directory to the temporary directory
local.get_temproot().chdir()


# %% GLOBALS

# %% HELPER DEFINITIONS

# %% FIXTURES

@pytest.fixture
def h5_data(request):
    """
    create dummy hdf5 test data file for testing PyContainer and H5NodeFilterProxy
    """
    import h5py as h5
    dummy_file = h5.File('hickle_core_{}.hdf5'.format(request.function.__name__),'w')
    filename = dummy_file.filename
    test_data = dummy_file.create_group("root_group")
    yield test_data
    dummy_file.close()
    
@pytest.fixture
def test_file_name(request):
    yield "{}.hkl".format(request.function.__name__)

# %% FUNCTION DEFINITIONS
        
def test_recursive_dump(h5_data,compression_kwargs):
    """
    test _dump function and that it properly calls itself recursively 
    """

    # check that dump function properly creates a list dataset and
    # sets appropriate values for 'type' and 'base_type' attributes
    data = simple_list = [1,2,3,4]
    with lookup.ReferenceManager.create_manager(h5_data) as memo:
        with lookup.LoaderManager.create_manager(h5_data) as loader:
            hickle._dump(data, h5_data, "simple_list",memo,loader,**compression_kwargs)
            dumped_data = h5_data["simple_list"]
            assert memo.resolve_type(dumped_data) == (data.__class__,b'list',False)
            assert np.all(dumped_data[()] == simple_list)
        
            # check that dump function properly creates a group representing
            # a dictionary and its keys and values and sets appropriate values
            # for 'type', 'base_type' and 'key_base_type' attributes
            data = {
                '12':12,
                (1,2,3):'hallo'
            }
            hickle._dump(data, h5_data, "some_dict",memo,loader,**compression_kwargs)
            dumped_data = h5_data["some_dict"]
            assert memo.resolve_type(dumped_data) == (data.__class__,b'dict',True)
        
            # check that the name of the resulting dataset for the first dict item
            # resembles double quoted string key and 'type', 'base_type 'key_base_type'
            # attributes the resulting dataset are set accordingly
            first_item = dumped_data['"12"']
            assert first_item[()] == 12 and first_item.attrs['key_base_type'] in (b'str','str')
            assert memo.resolve_type(first_item) == (data['12'].__class__,b'int',False)
            #assert first_item.attrs['base_type'] == b'int'
            #assert first_item.attrs['type'] == pickle.dumps(data['12'].__class__) 
            
            # check that second item is converted into key value pair group, that
            # the name of that group reads 'data0' and that 'type', 'base_type' and
            # 'key_base_type' attributes are set accordingly
            second_item = dumped_data.get("data0",None)
            if second_item is None:
                second_item = dumped_data["data1"]
            assert second_item.attrs['key_base_type'] in (b'key_value','key_value')
            assert memo.resolve_type(second_item) == (tuple,b'tuple',True)
            #assert second_item.attrs['type'] == pickle.dumps(tuple)
        
            # check that content of key value pair group resembles key and value of
            # second dict item
            key = second_item['data0']
            value = second_item['data1']
            assert np.all(key[()] == (1,2,3))
            # and key.attrs['base_type'] == b'tuple'
            assert memo.resolve_type(key) == (tuple,b'tuple',False)
            assert bytes(value[()]) == 'hallo'.encode('utf8')
            # and value.attrs['base_type'] == b'str'
            assert memo.resolve_type(value) == (str,b'str',False)
        
            # check that objects for which no loader has been registered or for which
            # available loader raises NotHicklable exception are handled by 
            # create_pickled_dataset function 
            def fail_create_dict(py_obj,h_group,name,**kwargs):
                raise helpers.NotHicklable("test loader shrugg")
            loader.types_dict.maps.insert(0,{dict:(fail_create_dict,*loader.types_dict[dict][1:])})
            memo_backup = memo.pop(id(data),None)
            with pytest.warns(lookup.SerializedWarning):
                hickle._dump(data, h5_data, "pickled_dict",memo,loader,**compression_kwargs)
            dumped_data = h5_data["pickled_dict"]
            assert bytes(dumped_data[()]) == pickle.dumps(data)
            loader.types_dict.maps.pop(0)
            memo[id(data)] = memo_backup
    
def test_recursive_load(h5_data,compression_kwargs):
    """
    test _load function and that it properly calls itself recursively 
    """

    # check that simple scalar value is properly restored on load from
    # corresponding dataset
    data = 42
    data_name = "the_answer"
    with lookup.ReferenceManager.create_manager(h5_data) as memo:
        with lookup.LoaderManager.create_manager(h5_data) as loader:
            hickle._dump(data, h5_data, data_name,memo,loader,**compression_kwargs)
            py_container = hickle.RootContainer(h5_data.attrs,b'hickle_root',hickle.RootContainer)
            hickle._load(py_container, data_name, h5_data[data_name],memo,loader)
            assert py_container.convert() == data
        
            # check that dict object is properly restored on load from corresponding group
            data = {'question':None,'answer':42}
            data_name = "not_formulated"
            hickle._dump(data, h5_data, data_name,memo,loader,**compression_kwargs)
            py_container = hickle.RootContainer(h5_data.attrs,b'hickle_root',hickle.RootContainer)
            hickle._load(py_container, data_name, h5_data[data_name],memo,loader)
            assert py_container.convert() == data
        
            
            # check that objects for which no loader has been registered or for which
            # available loader raises NotHicklable exception are properly restored on load
            # from corresponding copy protocol group or pickled data string 
            def fail_create_dict(py_obj,h_group,name,**kwargs):
                raise helpers.NotHicklable("test loader shrugg")
            loader.types_dict.maps.insert(0,{dict:(fail_create_dict,*loader.types_dict[dict][1:])})
            data_name = "pickled_dict"
            memo_backup = memo.pop(id(data),None)
            with pytest.warns(lookup.SerializedWarning):
                hickle._dump(data, h5_data, data_name,memo,loader,**compression_kwargs)
            hickle._load(py_container, data_name, h5_data[data_name],memo,loader)
            assert py_container.convert() == data
            loader.types_dict.maps.pop(0)
            memo[id(data)] = memo_backup

# %% ISSUE RELATED TESTS

def test_invalid_file(compression_kwargs):
    """ Test if trying to use a non-file object fails. """

    with pytest.raises(hickle.FileError):
        dump('test', (),**compression_kwargs)


def test_binary_file(test_file_name,compression_kwargs):
    """ Test if using a binary file works

    https://github.com/telegraphic/hickle/issues/123"""

    filename = test_file_name.replace(".hkl",".hdf5")
    with open(filename, "w") as f:
        with pytest.raises(hickle.FileError):
            hickle.dump(None, f,**compression_kwargs)
    with open(filename, "w+") as f:
        with pytest.raises(hickle.FileError):
            hickle.dump(None, f,**compression_kwargs)

    with open(filename, "wb") as f:
        with pytest.raises(hickle.FileError):
            hickle.dump(None, f,**compression_kwargs)

    with open(filename, "w+b") as f:
        hickle.dump(None, f,**compression_kwargs)


def test_file_open_close(test_file_name,h5_data,compression_kwargs):
    """ https://github.com/telegraphic/hickle/issues/20 """
    import h5py
    f = h5py.File(test_file_name.replace(".hkl",".hdf"), 'w')
    a = np.arange(5)

    dump(a, test_file_name,**compression_kwargs)
    dump(a, test_file_name,**compression_kwargs)

    dump(a, f, mode='w',**compression_kwargs)
    f.close()
    with pytest.raises(hickle.ClosedFileError):
        dump(a, f, mode='w',**compression_kwargs)
    h5_data.create_dataset('nothing',data=[])
    with pytest.raises(ValueError,match = r"Unable\s+to\s+create\s+group\s+\(name\s+already\s+exists\)"):
        dump(a,h5_data.file,path="/root_group",**compression_kwargs)


def test_hdf5_group(test_file_name,compression_kwargs):
    import h5py
    hdf5_filename = test_file_name.replace(".hkl",".hdf5")
    file = h5py.File(hdf5_filename, 'w')
    group = file.create_group('test_group')
    a = np.arange(5)
    dump(a, group,**compression_kwargs)
    file.close()

    a_hkl = load(hdf5_filename, path='/test_group')
    assert np.allclose(a_hkl, a)

    file = h5py.File(hdf5_filename, 'r+')
    group = file.create_group('test_group2')
    b = np.arange(8)

    dump(b, group, path='deeper/and_deeper',**compression_kwargs)
    file.close()

    with pytest.raises(ValueError):
        b_hkl = load(hdf5_filename, path='/test_group2/deeper_/and_deeper')
    b_hkl = load(hdf5_filename, path='/test_group2/deeper/and_deeper')
    assert np.allclose(b_hkl, b)

    file = h5py.File(hdf5_filename, 'r')
    b_hkl2 = load(file['test_group2'], path='deeper/and_deeper')
    assert np.allclose(b_hkl2, b)
    file.close()



def test_with_open_file(test_file_name,compression_kwargs):
    """
    Testing dumping and loading to an open file

    https://github.com/telegraphic/hickle/issues/92"""

    lst = [1]
    tpl = (1,)
    dct = {1: 1}
    arr = np.array([1])

    with h5py.File(test_file_name, 'w') as file:
        dump(lst, file, path='/lst',**compression_kwargs)
        dump(tpl, file, path='/tpl',**compression_kwargs)
        dump(dct, file, path='/dct',**compression_kwargs)
        dump(arr, file, path='/arr',**compression_kwargs)

    with h5py.File(test_file_name, 'r') as file:
        assert load(file, '/lst') == lst
        assert load(file, '/tpl') == tpl
        assert load(file, '/dct') == dct
        assert load(file, '/arr') == arr


def test_load(test_file_name,compression_kwargs):
    a = set([1, 2, 3, 4])
    b = set([5, 6, 7, 8])
    c = set([9, 10, 11, 12])
    z = (a, b, c)
    z = [z, z]
    z = (z, z, z, z, z)

    print("Original:")
    pprint(z)
    dump(z, test_file_name, mode='w',**compression_kwargs)

    print("\nReconstructed:")
    z = load(test_file_name)
    pprint(z)




def test_multi_hickle(test_file_name,compression_kwargs):
    """ Dumping to and loading from the same file several times

    https://github.com/telegraphic/hickle/issues/20"""

    a = {'a': 123, 'b': [1, 2, 4]}

    if os.path.exists(test_file_name):
        os.remove(test_file_name)
    dump(a, test_file_name, path="/test", mode="w",**compression_kwargs)
    dump(a, test_file_name, path="/test2", mode="r+",**compression_kwargs)
    dump(a, test_file_name, path="/test3", mode="r+",**compression_kwargs)
    dump(a, test_file_name, path="/test4", mode="r+",**compression_kwargs)

    load(test_file_name, path="/test")
    load(test_file_name, path="/test2")
    load(test_file_name, path="/test3")
    load(test_file_name, path="/test4")


def test_improper_attrs(test_file_name,compression_kwargs):
    """
    test for proper reporting missing mandatory attributes for the various
    supported file versions
    """

    # check that missing attributes which disallow to identify
    # hickle version are reported
    data = "my name? Ha I'm Nobody"
    dump(data,test_file_name,**compression_kwargs)
    manipulated = h5py.File(test_file_name,"r+")
    root_group = manipulated.get('/')
    root_group.attrs["VERSION"] = root_group.attrs["HICKLE_VERSION"]
    del root_group.attrs["HICKLE_VERSION"]
    manipulated.flush()
    with pytest.raises(
        ValueError,
        match= r"Provided\s+argument\s+'file_obj'\s+does\s+not\s+appear"
               r"\s+to\s+be\s+a\s+valid\s+hickle\s+file!.*"
    ):
        load(manipulated)


# %% MAIN SCRIPT
if __name__ == '__main__':
    """ Some tests and examples """
    from _pytest.fixtures import FixtureRequest
    from hickle.tests.conftest import compression_kwargs

    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_recursive_dump),)
    ):
        test_recursive_dump(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_recursive_load),)
    ):
        test_recursive_load(h5_root,keywords)
    for keywords in compression_kwargs(FixtureRequest(test_recursive_dump)):
        test_invalid_file(keywords)
    for filename,keywords in (
        ( test_file_name(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_binary_file),)
    ):
        test_binary_file(filename,keywords)
    for h5_root,filename,keywords in (
        ( h5_data(request),test_file_name(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_file_open_close),)
    ):
        test_file_open_close(h5_root,filename,keywords)
    for filename,keywords in (
        ( test_file_name(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_hdf5_group),)
    ):
        test_hdf5_group(filename,keywords)
    for filename,keywords in (
        ( test_file_name(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_with_open_file),)
    ):
        test_with_open_file(filename,keywords)
    for filename,keywords in (
        ( test_file_name(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_load),)
    ):
        test_load(filename,keywords)
    for filename,keywords in (
        ( test_file_name(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_multi_hickle),)
    ):
        test_multi_hickle(filename,keywords)
    for filename,keywords in (
        ( test_file_name(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_improper_attrs),)
    ):
        test_improper_attrs(filename,keywords)




import numpy as np
from astropy.units import Quantity
from astropy.coordinates import Angle, SkyCoord
from astropy.constants import Constant, EMConstant
from astropy.table import Table
from astropy.time import Time

from ..helpers import get_type_and_data
import six

def create_astropy_quantity(py_obj, h_group, call_id=0, **kwargs):
    """ dumps an astropy quantity

    Args:
        py_obj: python object to dump; should be a python type (int, float, bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the iterable.
    """
    # kwarg compression etc does not work on scalars
    d = h_group.create_dataset('data_%i' % call_id, data=py_obj.value,
                               dtype='float64')     #, **kwargs)
    d.attrs["type"] = [b'astropy_quantity']
    if six.PY3:
        unit = bytes(str(py_obj.unit), 'ascii')
    else:
        unit = str(py_obj.unit)
    d.attrs['unit'] = [unit]

def create_astropy_angle(py_obj, h_group, call_id=0, **kwargs):
    """ dumps an astropy quantity

    Args:
        py_obj: python object to dump; should be a python type (int, float, bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the iterable.
    """
    # kwarg compression etc does not work on scalars
    d = h_group.create_dataset('data_%i' % call_id, data=py_obj.value,
                               dtype='float64')     #, **kwargs)
    d.attrs["type"] = [b'astropy_angle']
    if six.PY3:
        unit = str(py_obj.unit).encode('ascii')
    else:
        unit = str(py_obj.unit)
    d.attrs['unit'] = [unit]

def create_astropy_skycoord(py_obj, h_group, call_id=0, **kwargs):
    """ dumps an astropy quantity

    Args:
        py_obj: python object to dump; should be a python type (int, float, bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the iterable.
    """
    # kwarg compression etc does not work on scalars
    lat = py_obj.data.lat.value
    lon = py_obj.data.lon.value
    dd = np.stack((lon, lat), axis=-1)

    d = h_group.create_dataset('data_%i' % call_id, data=dd,
                               dtype='float64')     #, **kwargs)
    d.attrs["type"] = [b'astropy_skycoord']
    if six.PY3:
        lon_unit = str(py_obj.data.lon.unit).encode('ascii')
        lat_unit = str(py_obj.data.lat.unit).encode('ascii')
    else:
        lon_unit = str(py_obj.data.lon.unit)
        lat_unit = str(py_obj.data.lat.unit)
    d.attrs['lon_unit'] = [lon_unit]
    d.attrs['lat_unit'] = [lat_unit]

def create_astropy_time(py_obj, h_group, call_id=0, **kwargs):
    """ dumps an astropy Time object

    Args:
        py_obj: python object to dump; should be a python type (int, float, bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the iterable.
    """

    # kwarg compression etc does not work on scalars
    data = py_obj.value
    dtype = str(py_obj.value.dtype)

    # Need to catch string times
    if '<U' in dtype:
        dtype = dtype.replace('<U', '|S')
        print(dtype)
        data = []
        for item in py_obj.value:
            data.append(str(item).encode('ascii'))

    d = h_group.create_dataset('data_%i' % call_id, data=data, dtype=dtype)     #, **kwargs)
    d.attrs["type"] = [b'astropy_time']
    if six.PY2:
        fmt   = str(py_obj.format)
        scale = str(py_obj.scale)
    else:
        fmt   = str(py_obj.format).encode('ascii')
        scale = str(py_obj.scale).encode('ascii')
    d.attrs['format'] = [fmt]
    d.attrs['scale']  = [scale]

def create_astropy_constant(py_obj, h_group, call_id=0, **kwargs):
    """ dumps an astropy constant

    Args:
        py_obj: python object to dump; should be a python type (int, float, bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the iterable.
    """
    # kwarg compression etc does not work on scalars
    d = h_group.create_dataset('data_%i' % call_id, data=py_obj.value,
                               dtype='float64')     #, **kwargs)
    d.attrs["type"]   = [b'astropy_constant']
    d.attrs["unit"]   = [str(py_obj.unit)]
    d.attrs["abbrev"] = [str(py_obj.abbrev)]
    d.attrs["name"]   = [str(py_obj.name)]
    d.attrs["reference"] = [str(py_obj.reference)]
    d.attrs["uncertainty"] = [py_obj.uncertainty]

    if py_obj.system:
        d.attrs["system"] = [py_obj.system]


def create_astropy_table(py_obj, h_group, call_id=0, **kwargs):
    """ Dump an astropy Table

    Args:
        py_obj: python object to dump; should be a python type (int, float, bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the iterable.
    """
    data = py_obj.as_array()
    d = h_group.create_dataset('data_%i' % call_id, data=data, dtype=data.dtype, **kwargs)
    d.attrs['type']  = [b'astropy_table']

    if six.PY3:
        colnames = [bytes(cn, 'ascii') for cn in py_obj.colnames]
    else:
        colnames = py_obj.colnames
    d.attrs['colnames'] = colnames
    for key, value in py_obj.meta.items():
     d.attrs[key] = value


def load_astropy_quantity_dataset(h_node):
    py_type, data = get_type_and_data(h_node)
    unit = h_node.attrs["unit"][0]
    q = Quantity(data, unit, copy=False)
    return q

def load_astropy_time_dataset(h_node):
    py_type, data = get_type_and_data(h_node)
    if six.PY3:
        fmt = h_node.attrs["format"][0].decode('ascii')
        scale = h_node.attrs["scale"][0].decode('ascii')
    else:
        fmt = h_node.attrs["format"][0]
        scale = h_node.attrs["scale"][0]
    q = Time(data, format=fmt, scale=scale)
    return q

def load_astropy_angle_dataset(h_node):
    py_type, data = get_type_and_data(h_node)
    unit = h_node.attrs["unit"][0]
    q = Angle(data, unit)
    return q

def load_astropy_skycoord_dataset(h_node):
    py_type, data = get_type_and_data(h_node)
    lon_unit = h_node.attrs["lon_unit"][0]
    lat_unit = h_node.attrs["lat_unit"][0]
    q = SkyCoord(data[..., 0], data[..., 1], unit=(lon_unit, lat_unit))
    return q

def load_astropy_constant_dataset(h_node):
    py_type, data = get_type_and_data(h_node)
    unit   = h_node.attrs["unit"][0]
    abbrev = h_node.attrs["abbrev"][0]
    name   = h_node.attrs["name"][0]
    ref    = h_node.attrs["reference"][0]
    unc    = h_node.attrs["uncertainty"][0]

    system = None
    if "system" in h_node.attrs.keys():
        system = h_node.attrs["system"][0]

    c = Constant(abbrev, name, data, unit, unc, ref, system)
    return c

def load_astropy_table(h_node):
    py_type, data = get_type_and_data(h_node)
    metadata = dict(h_node.attrs.items())
    metadata.pop('type')
    metadata.pop('colnames')

    if six.PY3:
        colnames = [cn.decode('ascii') for cn in h_node.attrs["colnames"]]
    else:
        colnames = h_node.attrs["colnames"]

    t = Table(data, names=colnames, meta=metadata)
    return t

def check_is_astropy_table(py_obj):
    return isinstance(py_obj, Table)

def check_is_astropy_quantity_array(py_obj):
    if isinstance(py_obj, Quantity) or isinstance(py_obj, Time) or \
       isinstance(py_obj, Angle) or isinstance(py_obj, SkyCoord):
        if py_obj.isscalar:
            return False
        else:
            return True
    else:
        return False


#####################
# Lookup dictionary #
#####################

class_register = [
    [Quantity, b'astropy_quantity', create_astropy_quantity, load_astropy_quantity_dataset,
     True, check_is_astropy_quantity_array],
    [Time,     b'astropy_time', create_astropy_time, load_astropy_time_dataset,
     True, check_is_astropy_quantity_array],
    [Angle,    b'astropy_angle', create_astropy_angle, load_astropy_angle_dataset,
     True, check_is_astropy_quantity_array],
    [SkyCoord, b'astropy_skycoord', create_astropy_skycoord, load_astropy_skycoord_dataset,
     True, check_is_astropy_quantity_array],
    [Constant, b'astropy_constant', create_astropy_constant, load_astropy_constant_dataset,
     True, None],
    [Table,    b'astropy_table',  create_astropy_table, load_astropy_table,
     True, check_is_astropy_table]
]



# encoding: utf-8
"""
# hickle.py

Created by Danny Price 2016-02-03.

Hickle is an HDF5 based clone of Pickle. Instead of serializing to a pickle
file, Hickle dumps to an HDF5 file. It is designed to be as similar to pickle
in usage as possible, providing a load() and dump() function.

## Notes

Hickle has two main advantages over Pickle:
1) LARGE PICKLE HANDLING. Unpickling a large pickle is slow, as the Unpickler
reads the entire pickle thing and loads it into memory. In comparison, HDF5
files are designed for large datasets. Things are only loaded when accessed.

2) CROSS PLATFORM SUPPORT. Attempting to unpickle a pickle pickled on Windows
on Linux and vice versa is likely to fail with errors like "Insecure string
pickle". HDF5 files will load fine, as long as both machines have
h5py installed.

3) pickle.dumps() and pickle.loads() functions can be mimicked by passing
a BytesIO type to hickle.dump() or hickle.load() function and setting the
filename parameter to a non empty string.

    hicklestring = BytesIO()
    hickle.dump(my_data,hicklestring,mode='w',filename='<string>')

    loaded_data = hickle.load(hicklestring,mode='r',filename='<string>')

"""


# %% IMPORTS
# Built-in imports
import io
import sys
import warnings
import types
import functools as ft

# Package imports
import pickle
import h5py as h5
import numpy as np

#whished it would not be necessary but sometimes garbage collector
#may kick in while trying to close file. Causing ValueError in close
#to prevent check if collectoin is necessary after flushing file
import gc

# hickle imports
from hickle import __version__
from .helpers import (
    PyContainer, NotHicklable, nobody_is_my_name, ToDoError
)
from .fileio import ClosedFileError, FileError, file_opener
from .lookup import (
    #hkl_types_dict, hkl_container_dict, load_loader, load_legacy_loader ,
    create_pickled_dataset, load_nothing, fix_lambda_obj_type,ReferenceManager,
    LoaderManager, RecoverGroupContainer, recover_custom_dataset
)


# All declaration
__all__ = ['dump', 'load']


# %% FUNCTION DEFINITIONS


###########
# DUMPERS #
###########

def _dump(py_obj, h_group, name, memo, loader,attrs={} , **kwargs):
    """ Dump a python object to a group within an HDF5 file.

    This function is called recursively by the main dump() function.

    Parameters
    ----------
    py_obj (object):
         python object to dump.

    h_group (h5.File.group):
         group to dump data into.

    name (str):
         name of resulting hdf5 group or dataset 

    memo (ReferenceManager):
        the ReferenceManager object responsible for handling all
        object and type memoisation related issues

    attrs (dict):
        additional attributes to be stored along with the resulting
        hdf5 group or hdf5 dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function
    """

    py_obj_id = id(py_obj)
    py_obj_ref = memo.get(py_obj_id,None)
    if py_obj_ref is not None:

        # py_object already dumped to hdf5 file store a reference to it instead
        # instead of dumping it again.
        #
        # Note: reference dataset share their base_type and py_obj_type with the
        #       referenced h5py.Group or h5py.Dataset. On load their h5py.ref_dtype type 
        #       dtype is used to distinguish them from datasets hosting pickled data.
        h_link = h_group.create_dataset(name,data = py_obj_ref[0].ref,dtype = h5.ref_dtype)
        h_link.attrs.update(attrs)
        return

    # Check if loader has already been loaded for the provided py_obj and 
    # retrieve the most appropriate method for creating the corresponding
    # representation within HDF5 file
    py_obj_type, (create_dataset, base_type,memoise) = loader.load_loader(py_obj.__class__)
    try:
        h_node,h_subitems = create_dataset(py_obj, h_group, name, **kwargs)
    except NotHicklable:
        h_node,h_subitems = create_pickled_dataset(py_obj, h_group, name, reason = str(NotHicklable), **kwargs)
    else:
        # store base_type and type unless py_obj had to be pickled by create_pickled_dataset
        memo.store_type(h_node,py_obj_type,base_type,**kwargs)

    # add additional attributes and prevent modification of 'type' attribute
    h_node.attrs.update((name,attr) for name,attr in attrs.items() if name != 'type' )

    # if py_object shall be memoised to properly represent multiple references
    # to it in HDF5 file store it along with created h_node in the memo dictionary.
    # remembering the py_object along with the h_node ensures that py_object_id
    # which represents the memory address of py_obj refers to py_obj until the
    # whole structure is stored within hickle file.
    if memoise:
        memo[py_obj_id] = (h_node,py_obj)

    # loop through list of all sub items and recursively dump them
    # to HDF5 file
    for h_subname,py_subobj,h_subattrs,sub_kwargs in h_subitems:
        _dump(py_subobj,h_node,h_subname,memo,loader,h_subattrs,**sub_kwargs)


def dump(py_obj, file_obj, mode='w', path='/',*,filename = None,options = {},**kwargs):
    """
    Write a hickled representation of `py_obj` to the provided `file_obj`.

    Parameters
    ----------
    py_obj (object):
        Python object to hickle to HDF5.

    file_obj (file, file-like, h5py.File, str, (file,str),{'file':file,'name':str} ):
        File to open for dumping or loading purposes.
        str:
            the path of the HDF5-file that must be used.
        ~h5py.Group:
             the group (or file) in an open HDF5-file that must be used.
        file, file-like: 
            file or like object which provides `read`, `seek`, `tell` and write methods
        tuple:
            two element tuple with the first being the file or file like object
            to dump to and the second the filename to be used instead of 'filename'
            parameter
        dict:
            dictionary with 'file' and 'name' items

    mode (str): optional
        string indicating how the file shall be opened. For details see Python `open`.
        
        Note: The 'b' flag is optional as all files are and have to be opened in
            binary mode.

    path (str): optional
        Path within HDF5-file or group to dump to/load from.
    
    filename (str): optional
        The name of the file. Ignored when f is `str` or `h5py.File` object.

    options (dict): optional
        Each entry in this dict modifies how hickle dumps data to file.
        For example 
            { custom = True }
        would enforce use of custom loaders on all classes
        registered with this kind of loader.
            { custom = False }
        would disable custom loaders for dumped data even if
        globally turned on. More options may follow.

    kwargs : keyword arguments
        Additional keyword arguments that must be provided to the
        :meth:`~h5py.Group.create_dataset` method. For example compression=True

    Raises
    ------
    CloseFileError:
        If passed h5py.File, h5py.Group or h5py.Dataset object is not
        accessible. This in most cases indicates that underlying HDF5
        was closed or if file or file or file-like object has already been 
        closed.

    FileError
        If passed file or file-like object is not opened for reading or
        in addition for writing in case mode corresponds to any
        of 'w', 'w+', 'x', 'x+' or a.

    ValueError:
        If anything else than str, bytes or None specified for filename
        or for mode is anything else specified than 'w','w+','x','x+','r','r+','a'
        or contains any optional open flag other than 'b'
    """

    # Open the file
    h5f, path, close_flag = file_opener(file_obj, path, mode,filename)
    try:

        # Log which version of python was used to generate the hickle file
        pv = sys.version_info
        py_ver = "%i.%i.%i" % (pv[0], pv[1], pv[2])

        h_root_group = h5f.get(path,None)
        if h_root_group is None:
            h_root_group = h5f.create_group(path)
        elif h_root_group.items():
            raise ValueError("Unable to create group (name already exists)")

        h_root_group.attrs["HICKLE_VERSION"] = __version__
        h_root_group.attrs["HICKLE_PYTHON_VERSION"] = py_ver

        with LoaderManager.create_manager(h_root_group,False,options) as loader:
            with ReferenceManager.create_manager(h_root_group) as memo:
                _dump(py_obj, h_root_group,'data', memo ,loader,**kwargs)
    finally:
        # Flush the the h5py.File  and close it if it was opened by hickle.
        h5f.flush()

        # disable python garbage collector while closing to prevent Unrecognized
        # typecode ValueError caused by h5py objects collected to earlay most 
        # persistently observed with h5py 2.10 in python 3.7 on 32 bit windows
        gc.disable()
        if close_flag:
            h5f.close()
        gc.enable()

###########
# LOADERS #
###########

class RootContainer(PyContainer):
    """
    PyContainer representing the whole HDF5 file
    """

    __slots__ = ()
    def convert(self):
        return self._content[0]


def load(file_obj, path='/', safe=True, filename = None):
    """
    Load the Python object stored in `file_obj` at `path` and return it.

    Parameters
    ----------

    file_obj (file, file-like, h5py.File, str, (file,str),{'file':file,'name':str} ):
        File to open for dumping or loading purposes.
        str:
            the path of the HDF5-file that must be used.
        ~h5py.Group:
             the group (or file) in an open HDF5-file that must be used.
        file, file-like: 
            file or like object which provides `read`, `seek`, `tell` and write methods
        tuple:
            two element tuple with the first being the file or file like object
            to dump to and the second the filename to be used instead of 'filename'
            parameter
        dict:
            dictionary with 'file' and 'name' items

    
    path (str): optional
            Path within HDF5-file or group to dump to/load from.

    safe (bool): optional
        Disable automatic depickling of arbitrary python objects.
        DO NOT set this to False unless the file is from a trusted source.
        (See https://docs.python.org/3/library/pickle.html for an explanation)

        Note: ignored when loading hickle 4.x and newer files

        
    filename (str): optional
        The name of the file. Ignored when f is `str` or `h5py.File` object.

    Returns
    -------
    py_obj : object
        The unhickled Python object.

    Raises
    ------
    CloseFileError:
        If passed h5py.File, h5py.Group or h5py.Dataset object is not
        accessible. This in most cases indicates that underlying HDF5
        was closed or if file or file or file-like object has already been 
        closed.

    FileError
        If passed file or file-like object is not opened for reading

    ValueError:
        If anything else than str, bytes or None specified for filename
    """

    # Try to read the provided file_obj as a hickle file
    h5f, path, close_flag = file_opener(file_obj, path, 'r', filename)
    try:
        h_root_group = h5f.get(path,None) # only used by v4
        if not isinstance(h_root_group,h5.Group):
            raise FileError("file '{}': path '{}' not existing".format(h5f.filename,path))

        # Define attributes h_root_group must have
        v3_attrs = ['CLASS', 'VERSION', 'PYTHON_VERSION']
        v4_attrs = ['HICKLE_VERSION', 'HICKLE_PYTHON_VERSION']

        # Check if the proper attributes for v3 loading are available
        if all(map(h5f.attrs.get, v3_attrs)):
            # Check if group attribute 'CLASS' has value 'hickle
            if(h5f.attrs['CLASS'] not in ( b'hickle','hickle')):  # pragma: no cover
                # If not, raise error
                raise AttributeError("HDF5-file attribute 'CLASS' does not "
                                     "have value 'hickle'!")

            # Obtain version with which the file was made
            try:
                major_version = int(h5f.attrs['VERSION'][0])

            # If this cannot be done, then this is not a v3 file
            except Exception:  # pragma: no cover
                raise Exception("This file does not appear to be a hickle v3 "
                                "file.")

            # Else, if the major version is not 3, it is not a v3 file either
            else:
                if(major_version != 3):  # pragma: no cover
                    raise Exception("This file does not appear to be a hickle "
                                    "v3 file.")

            # Load file
            from hickle import legacy_v3
            warnings.warn("Input argument 'file_obj' appears to be a file made"
                          " with hickle v3. Using legacy load...")
            return(legacy_v3.load(file_obj, path, safe))

        # Else, check if the proper attributes for v4 loading are available
        if all(map(h_root_group.attrs.get, v4_attrs)):
            # Load file
            py_container = RootContainer(h_root_group.attrs,b'document_root',RootContainer)
            pickle_loads = pickle.loads
            hickle_version = h_root_group.attrs["HICKLE_VERSION"].split('.')
            if int(hickle_version[0]) == 4:
                # hickle 4.x file activate if legacy load fixes for 4.x
                # eg. pickle of versions < 3.8 do not prevent dumping of lambda functions
                # even though stated otherwise in documentation. Activate workarounds
                # just in case issues arise. Especially as corresponding lambdas in
                # load_numpy are not needed anymore and thus have been removed.
                with LoaderManager.create_manager(h_root_group,True) as loader:
                    with ReferenceManager.create_manager(h_root_group,fix_lambda_obj_type) as memo:
                        _load(py_container, 'data',h_root_group['data'],memo,loader) #load_loader = load_legacy_loader)
                return py_container.convert()
            # 4.1.x file and newer
            with LoaderManager.create_manager( h_root_group,False) as loader:
                with ReferenceManager.create_manager(h_root_group,pickle_loads) as memo:
                    _load(py_container, 'data',h_root_group['data'],memo,loader) #load_loader = load_loader)
            return py_container.convert()

        # Else, raise error
        raise FileError("HDF5-file does not have the proper attributes!")

    # If this fails, raise error and provide user with caught error message
    except Exception as error:
        raise ValueError("Provided argument 'file_obj' does not appear to be a valid hickle file! (%s)" % (error),error) from error
    finally:
        # Flush the h5py.File and close it lif it was opened by hickle.
        h5f.flush()

        # disable python garbage collector while closing to prevent Unrecognized
        # typecode ValueError caused by h5py objects collected to earlay most 
        # persistently observed with h5py 2.10 in python 3.7 on 32 bit windows
        gc.disable()
        if close_flag:
            h5f.close()
        gc.enable()



def _load(py_container, h_name, h_node,memo,loader): #load_loader = load_loader):
    """ Load a hickle file

    Recursive function to load hdf5 data into a PyContainer()

    Parameters
    ----------
    py_container (PyContainer):
        Python container to load data into

    h_name (str):
        the name of the resulting h5py.Group or h5py.Dataset

    h_node (h5py.Group, h5py.Dataset):
        h5py.Group or h5py.Dataset to restore data from.

    memo (ReferenceManager):
        the ReferenceManager object responsible for handling all object
        and type memoisation related issues

    loader (LoaderManager):
        the LoaderManager object managing the loaders required to properly
        restore the content of h_node and append it to py_container.
    """

    # if h_node has already been loaded cause a reference to it was encountered
    # earlier directly append it to its parent container and return
    node_ref = memo.get(h_node.id,h_node)
    if node_ref is not h_node:
        py_container.append(h_name,node_ref,h_node.attrs)
        return

    # load the type information of node.
    py_obj_type,base_type,is_container = memo.resolve_type(h_node)
    py_obj_type,(_,_,memoise) = loader.load_loader(py_obj_type,base_type=base_type)
    
    if is_container:
        # Either a h5py.Group representing the structure of complex objects or
        # a h5py.Dataset representing a h5py.Reference to the node of an object
        # referred to from multiple places within the object structure on dump 
        # is to be restored.
        # If no appropriate PyContainer is available use RecoverGroupContainer
        # instead to at least recover its contained data

        py_container_class = loader.hkl_container_dict.get(base_type,RecoverGroupContainer)
        py_subcontainer = py_container_class(h_node.attrs,base_type,py_obj_type)
    
        for h_key,h_subnode in py_subcontainer.filter(h_node):
            _load(py_subcontainer, h_key, h_subnode, memo ,loader)

        # finalize sub item
        sub_data = py_subcontainer.convert()
        py_container.append(h_name,sub_data,h_node.attrs)
    else:

        # must be a dataset load it and append to parent container.
        # In case no appropriate loader could be found use recover_custom_dataset
        # instead to at least recover the contained data
        load_fn = loader.hkl_types_dict.get(base_type, recover_custom_dataset)
        sub_data = load_fn(h_node,base_type,py_obj_type)
        py_container.append(h_name,sub_data,h_node.attrs)

    # store loaded object for properly restoring additional references to it
    if memoise:
        memo[h_node.id] = sub_data




# encoding: utf-8
"""
#helpers.py

Contains functions, classes and constants to be used
by all components of hickle including the loader modules
"""

# %% IMPORTS
# Built-in imports
import collections
import h5py as h5
import functools as ft

# Package imports


# %% EXCEPTION DEFINITIONS

nobody_is_my_name = ()

class NotHicklable(Exception):
    """
    object can not be mapped to proper hickle HDF5 file structure and
    thus shall be converted to pickle string before storing.
    """

class ToDoError(Exception):     # pragma: no cover
    """ An exception raised for non-implemented functionality"""
    def __str__(self):
        return "Error: this functionality hasn't been implemented yet."

# %% CLASS DEFINITIONS

class PyContainer():
    """
    Abstract base class for all PyContainer classes acting as proxy between
    h5py.Group and python object represented by the content of the h5py.Group.
    Any container type object as well as complex objects are represented
    in a tree like structure in the HDF5. PyContainer type objects ensure to
    properly map these structure when converting it into the corresponding
    python object structure.

    Parameters
    ----------
    h5_attrs (h5py.AttributeManager):
        attributes defined on h5py.Group object represented by this PyContainer

    base_type (bytes):
        the basic type used for representation in the HDF5 file

    object_type:
        type of Python object to be restored. May be used by PyContainer.convert
        to convert loaded Python object into final one.
        
    Attributes
    ----------
    base_type (bytes):
        the basic type used for representation on the HDF5 file

    object_type:
        type of Python object to be restored. Dependent upon container may
        be used by PyContainer.convert to convert loaded Python object into
        final one.
        
    """

    __slots__ = ("base_type", "object_type", "_h5_attrs", "_content","__dict__" )

    def __init__(self, h5_attrs, base_type, object_type, _content = None):
        """
        Parameters (protected):
        -----------------------
        _content (default: list):
            container to be used to collect the Python objects representing
            the sub items or the state of the final Python object. Shall only
            be set by derived PyContainer classes and not be set when default
            list container shall be used.

        """
        # the base type used to select this PyContainer
        self.base_type = base_type
        # class of python object represented by this PyContainer
        self.object_type = object_type
        # the h5_attrs structure of the h5_group to load the object_type from
        # can be used by the append and convert methods to obtain more
        # information about the container like object to be restored
        self._h5_attrs = h5_attrs
        # intermediate list, tuple, dict, etc. used to collect and store the sub items
        # when calling the append method
        self._content = _content if _content is not None else []

    def filter(self, h_parent):
        """
        PyContainer type child classes may overload this generator function
        to filter and preprocess the content of h_parent h5py.Group content 
        to ensure it can be properly processed by recursive calls to
        hickle._load function.

        Per default yields from h_parent.items(). 

        For examples see: 
            hickle.lookup.ExpandReferenceContainer.filter
            hickle.loaders.load_scipy.SparseMatrixContainer.filter
        """
        yield from h_parent.items()
 
    def append(self, name, item, h5_attrs):
        """
        adds the passed item to the content of this container.
       
        Parameters
        ----------
        name (string):
            the name of the h5py.Dataset or h5py.Group sub item was loaded from

        item:
            the Python object of the sub item

        h5_attrs:
            attributes defined on h5py.Group or h5py.Dataset sub item
            was loaded from.
        """
        self._content.append(item)

    def convert(self):
        """
        creates the final object and populates it with the items stored in the _content
        attribute. 

        Note: Must be implemented by the derived PyContainer child classes

        Returns
        -------
        py_obj:
            The final Python object loaded from file

        
        """
        raise NotImplementedError("convert method must be implemented")


class H5NodeFilterProxy():
    """
    Proxy class which allows to temporarily modify the content of h5_node.attrs
    attribute. Original attributes of underlying h5_node are left unchanged.
    
    Parameters
    ----------
    h5_node:
        node for which attributes shall be replaced by a temporary value
    """

    __slots__ = ('_h5_node','attrs','__dict__')

    def __init__(self,h5_node):
        # the h5py.Group or h5py.Dataset the attributes should temporarily
        # be modified.
        self._h5_node = h5_node
        # the temporarily modified attributes structure
        super().__setattr__( 'attrs', collections.ChainMap({}, h5_node.attrs))

    def __getattribute__(self, name):
        # for attrs and wrapped _h5_node return local copy. Any other request
        # redirect to wrapped _h5_node
        if name in {"attrs", "_h5_node"}:
            return super(H5NodeFilterProxy,self).__getattribute__(name)
        _h5_node = super(H5NodeFilterProxy,self).__getattribute__('_h5_node')
        return getattr(_h5_node, name)
        
    def __setattr__(self, name, value):
        # if wrapped _h5_node and attrs shall be set store value on local attributes
        # otherwise pass on to wrapped _h5_node
        if name in {'_h5_node'}:
            super().__setattr__(name, value)
            return
        if name in {'attrs'}: # pragma: no cover
            raise AttributeError('attribute is read-only')
        _h5_node = super().__getattribute__('_h5_node')
        setattr(_h5_node, name, value)    

    def __getitem__(self, *args, **kwargs):
        _h5_node = super().__getattribute__('_h5_node')
        return _h5_node.__getitem__(*args, **kwargs)
    # TODO as needed add more function like __getitem__ to fully proxy h5_node
    # or consider using metaclass __getattribute__ for handling special methods


class no_compression(dict):
    """
    named dict comprehension which temporarily removes any compression or
    data filter related argument from the passed iterable. 
    """

    # list of keyword parameters to filter
    __filter_keys__ = {
        "compression", "shuffle", "compression_opts", "chunks", "fletcher32", "scaleoffset"
    }

    def __init__(self, mapping):
        super().__init__((
            (key,value)
            for key,value in ( mapping.items() if isinstance(mapping,dict) else mapping )
            if key not in no_compression.__filter_keys__
        ))
        

# %% FUNCTION DEFINITIONS

def not_dumpable( py_obj, h_group, name, **kwargs): # pragma: no cover
    """
    create_dataset method attached to loader of dummy py_object which is used to
    mimic PyContainer class for  groups in legacy hickle 4.x file. 
        
    Raises
    ------
    RuntimeError:
        in any case as this function shall never be called    
    """

    raise RuntimeError("types defined by loaders not dump able")

def convert_str_attr(attrs,name,*,encoding='utf8'):
    return attrs[name].decode(encoding)

def convert_str_list_attr(attrs,name,*,encoding='utf8'):
    return [ value.decode(encoding) for value in attrs[name]]
  

if h5.version.version_tuple[0] >= 3: # pragma: no cover
    load_str_list_attr_ascii = load_str_list_attr = h5.AttributeManager.get
    load_str_attr_ascii = load_str_list_attr = h5.AttributeManager.get

else: # pragma: no cover
    load_str_list_attr_ascii = ft.partial(convert_str_list_attr,encoding='ascii')
    load_str_list_attr = convert_str_list_attr
    load_str_attr_ascii = ft.partial(convert_str_attr,encoding='ascii')
    load_str_attr = convert_str_attr




#! /usr/bin/env python
# encoding: utf-8
"""
# test_load_builtins

Unit tests for hickle module -- builtins loader.

"""

import pytest
import collections
import itertools

# %% IMPORTS
# Package imports
import h5py as h5
import numpy as np
from py.path import local

# hickle imports
import hickle.loaders.load_builtins as load_builtins
import hickle.helpers as helpers


# Set current working directory to the temporary directory
local.get_temproot().chdir()


# %% TEST DATA

dummy_data = (1,2,3)

# %% FIXTURES

@pytest.fixture
def h5_data(request):
    """
    create dummy hdf5 test data file providing parent group
    hosting created datasets and groups. Name of test function
    is included in filename
    """
    dummy_file = h5.File('load_builtins_{}.hdf5'.format(request.function.__name__),'w')
    filename = dummy_file.filename
    test_data = dummy_file.create_group("root_group")
    yield test_data
    dummy_file.close()


# %% FUNCTION DEFINITIONS

def test_scalar_dataset(h5_data,compression_kwargs):
    """
    tests creation and loading of datasets for scalar values
    """

    # check that scalar value is properly handled
    floatvalue = 5.2
    h_dataset,subitems= load_builtins.create_scalar_dataset(floatvalue,h5_data,"floatvalue",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and h_dataset[()] == floatvalue
    assert not [ item for item in subitems ]
    assert load_builtins.load_scalar_dataset(h_dataset,b'float',float) == floatvalue

    # check that integer value less than 64 bit is stored as int
    intvalue = 11
    h_dataset,subitems = load_builtins.create_scalar_dataset(intvalue,h5_data,"intvalue",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and h_dataset[()] == intvalue
    assert not [ item for item in subitems ]
    assert load_builtins.load_scalar_dataset(h_dataset,b'int',int) == intvalue

    # check that integer larger than 64 bit is stored as ASCII byte string
    non_mappable_int = int(2**65)
    h_dataset,subitems = load_builtins.create_scalar_dataset(non_mappable_int,h5_data,"non_mappable_int",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset)
    assert bytearray(h_dataset[()]) == str(non_mappable_int).encode('utf8')
    assert not [ item for item in subitems ]
    assert load_builtins.load_scalar_dataset(h_dataset,b'int',int) == non_mappable_int

    # check that integer larger than 64 bit is stored as ASCII byte string
    non_mappable_neg_int = -int(-2**63-1)
    h_dataset,subitems = load_builtins.create_scalar_dataset(non_mappable_neg_int,h5_data,"non_mappable_neg_int",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset)
    assert bytearray(h_dataset[()]) == str(non_mappable_neg_int).encode('utf8')
    assert not [ item for item in subitems ]
    assert load_builtins.load_scalar_dataset(h_dataset,b'int',int) == non_mappable_neg_int

def test_load_hickle_4_0_X_string(h5_data):
    string_data = "just test me as utf8 string"
    bytes_data = string_data.encode('utf8')
    if h5.version.version_tuple[0] >= 3:
        utf_entry = h5_data.create_dataset('utf_entry',data = string_data)#,dtype = 'U{}'.format(len(string_data)))
        bytes_entry = h5_data.create_dataset('bytes_entry',data = bytes_data,dtype = 'S{}'.format(len(bytes_data)))
    else:
        utf_entry = h5_data.create_dataset('utf_entry',data = string_data)
        bytes_entry = h5_data.create_dataset('bytes_entry',data = bytes_data)
    assert load_builtins.load_hickle_4_x_string(utf_entry,b'str',str) == string_data
    bytes_entry.attrs['str_type'] = b'str'
    assert load_builtins.load_hickle_4_x_string(bytes_entry,b'str',str) == string_data
    object_entry = h5_data.create_dataset('utf_h5py2_entry',data = string_data,dtype = np.dtype('O',metadata={'vlen':bytes}))
    assert load_builtins.load_hickle_4_x_string(object_entry,b'str',bytes) == bytes_data
    
    
def test_non_dataset(h5_data,compression_kwargs):
    """
     that None value is properly stored
    """
    h_dataset,subitems = load_builtins.create_none_dataset(None,h5_data,"None_value",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and h_dataset.shape is None and h_dataset.dtype == 'V1'
    assert not [ item for item in subitems ]
    assert load_builtins.load_none_dataset(h_dataset,b'None',None.__class__) is None


def test_listlike_dataset(h5_data,compression_kwargs):
    """
    test storing and loading of list like data 
    """

    # check that empty tuple is stored properly
    empty_tuple = ()
    h_dataset,subitems = load_builtins.create_listlike_dataset(empty_tuple, h5_data, "empty_tuple",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and h_dataset.size is None
    assert not subitems and iter(subitems)
    assert load_builtins.load_list_dataset(h_dataset,b'tuple',tuple) == empty_tuple

    # check that string data is stored properly stored as array of bytes
    # which supports compression
    stringdata = "string_data"
    h_dataset,subitems = load_builtins.create_listlike_dataset(stringdata, h5_data, "string_data",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not [ item for item in subitems ]
    assert bytearray(h_dataset[()]).decode("utf8") == stringdata
    assert h_dataset.attrs["str_type"] in ('str',b'str')
    assert load_builtins.load_list_dataset(h_dataset,b'str',str) == stringdata

    # check that byte string is properly stored as array of bytes which
    # supports compression
    bytesdata = b'bytes_data'
    h_dataset,subitems = load_builtins.create_listlike_dataset(bytesdata, h5_data, "bytes_data",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not [ item for item in subitems ]
    assert bytes(h_dataset[()]) == bytesdata
    assert h_dataset.attrs["str_type"] in ('bytes',b'bytes')
    assert load_builtins.load_list_dataset(h_dataset,b'bytes',bytes) == bytesdata

    # check that string dataset created by hickle 4.0.x is properly loaded 
    # utilizing numpy.array method. Mimic dumped data
    h_dataset = h5_data.create_dataset("legacy_np_array_bytes_data",data=np.array(stringdata.encode('utf8')))
    h_dataset.attrs['str_type'] = b'str'
    assert load_builtins.load_list_dataset(h_dataset,b'str',str) == stringdata

    # check that list of single type is stored as dataset of same type
    homogenous_list = [ 1, 2, 3, 4, 5, 6]
    h_dataset,subitems = load_builtins.create_listlike_dataset(homogenous_list,h5_data,"homogenous_list",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not [ item for item in subitems ]
    assert h_dataset[()].tolist() == homogenous_list and h_dataset.dtype == int
    assert load_builtins.load_list_dataset(h_dataset,b'list',list) == homogenous_list

    # check that list of different scalar types for which a least common type exists
    # is stored using a dataset 
    mixed_dtype_list = [ 1, 2.5, 3.8, 4, 5, 6]
    h_dataset,subitems = load_builtins.create_listlike_dataset(mixed_dtype_list,h5_data,"mixed_dtype_list",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not [ item for item in subitems ]
    assert h_dataset[()].tolist() == mixed_dtype_list and h_dataset.dtype == float
    assert load_builtins.load_list_dataset(h_dataset,b'list',list) == mixed_dtype_list

    # check that list containing non scalar objects is converted into group
    # further check that for groups representing list the index of items is either
    # provided via item_index attribute or can be read from name of item
    not_so_homogenous_list = [ 1, 2, 3, [4],5 ,6 ]
    h_dataset,subitems = load_builtins.create_listlike_dataset(not_so_homogenous_list,h5_data,"not_so_homogenous_list",**compression_kwargs)
    assert isinstance(h_dataset,h5.Group)
    item_name = "data{:d}"
    index = -1 
    loaded_list = load_builtins.ListLikeContainer(h_dataset.attrs,b'list',list)
    subitems1,subitems2 = itertools.tee(subitems,2)
    index_from_string = load_builtins.ListLikeContainer(h_dataset.attrs,b'list',list)
    for index,(name,item,attrs,kwargs) in enumerate(iter(subitems1)):
        assert item_name.format(index) == name and item == not_so_homogenous_list[index]
        assert attrs == {"item_index":index} and kwargs == compression_kwargs
        if isinstance(item,list):
            item_dataset,_ = load_builtins.create_listlike_dataset(item,h_dataset,name,**compression_kwargs)
        else:
            item_dataset = h_dataset.create_dataset(name,data = item)
        item_dataset.attrs.update(attrs)
        loaded_list.append(name,item,item_dataset.attrs)
        index_from_string.append(name,item,{})
    assert index + 1 == len(not_so_homogenous_list)
    assert loaded_list.convert() == not_so_homogenous_list
    assert index_from_string.convert() == not_so_homogenous_list

    # check that list groups which do not provide num_items attribute 
    # are automatically expanded to properly cover the highest index encountered
    # for any of the list items.
    no_num_items = {key:value for key,value in h_dataset.attrs.items() if key != "num_items"}
    no_num_items_container = load_builtins.ListLikeContainer(no_num_items,b'list',list)
    for index,(name,item,attrs,kwargs) in enumerate(iter(subitems2)):
        assert item_name.format(index) == name and item == not_so_homogenous_list[index]
        assert attrs == {"item_index":index} and kwargs == compression_kwargs
        item_dataset = h_dataset.get(name,None)
        no_num_items_container.append(name,item,{})
    assert index + 1 == len(not_so_homogenous_list)
    assert no_num_items_container.convert() == not_so_homogenous_list

    # check that list the first of which is not a scalar is properly mapped
    # to a group. Also check that ListLikeContainer.append raises exception
    # in case neither item_index is provided nor an index value can be parsed
    # from the tail of its name. Also check that ListLikeContainer.append
    # raises exception in case value for item_index already has been loaded
    object_list = [ [4, 5 ] ,6, [ 1, 2, 3 ] ]
    h_dataset,subitems = load_builtins.create_listlike_dataset(object_list,h5_data,"object_list",**compression_kwargs)
    assert isinstance(h_dataset,h5.Group)
    item_name = "data{:d}"
    wrong_item_name = item_name + "_ni"
    index = -1
    loaded_list = load_builtins.ListLikeContainer(h_dataset.attrs,b'list',list)
    index_from_string = load_builtins.ListLikeContainer(h_dataset.attrs,b'list',list)
    for index,(name,item,attrs,kwargs) in enumerate(iter(subitems)):
        assert item_name.format(index) == name and item == object_list[index]
        assert attrs == {"item_index":index} and kwargs == compression_kwargs
        if isinstance(item,list):
            item_dataset,_ = load_builtins.create_listlike_dataset(item,h_dataset,name,**compression_kwargs)
        else:
            item_dataset = h_dataset.create_dataset(name,data = item)
        item_dataset.attrs.update(attrs)
        loaded_list.append(name,item,item_dataset.attrs)
        with pytest.raises(KeyError,match = r"List\s+like\s+item name\s+'\w+'\s+not\s+understood"):
            index_from_string.append(wrong_item_name.format(index),item,{})
        # check that previous error is not triggered when
        # legacy 4.0.x loader injects the special value helpers.nobody_is_my_name which
        # is generated by load_nothing function. this is for example used as load method
        # for legacy 4.0.x np.masked.array objects where the mask is injected in parallel
        # in the root group of the corresponding values data set. By silently ignoring
        # this special value returned by load_nothing it can be assured that for example
        # mask datasets of numpy.masked.array objects hickup the loader.
        index_from_string.append(wrong_item_name.format(index),helpers.nobody_is_my_name,{})
        if index < 1:
            continue
        with pytest.raises(IndexError, match = r"Index\s+\d+\s+already\s+set"):
            loaded_list.append(name,item,{"item_index":index-1})
    assert index + 1 == len(object_list)

    # assert that list of strings where first string has length 1 is properly mapped
    # to group
    string_list = test_set = ['I','confess','appriciate','hickle','times']
    h_dataset,subitems = load_builtins.create_listlike_dataset(string_list,h5_data,"string_list",**compression_kwargs)
    assert isinstance(h_dataset,h5.Group)
    item_name = "data{:d}"
    index = -1 
    loaded_list = load_builtins.ListLikeContainer(h_dataset.attrs,b'list',list)
    index_from_string = load_builtins.ListLikeContainer(h_dataset.attrs,b'list',list)
    for index,(name,item,attrs,kwargs) in enumerate(iter(subitems)):
        assert item_name.format(index) == name and item == string_list[index]
        assert attrs == {"item_index":index} and kwargs == compression_kwargs
        item_dataset = h_dataset.create_dataset(name,data = item)
        item_dataset.attrs.update(attrs)
        loaded_list.append(name,item,item_dataset.attrs)
        index_from_string.append(name,item,{})
    assert index + 1 == len(string_list)
    assert loaded_list.convert() == string_list
    assert index_from_string.convert() == string_list

    # assert that list which contains numeric values and strings is properly mapped
    # to group
    mixed_string_list = test_set = [12,2.8,'I','confess','appriciate','hickle','times']
    h_dataset,subitems = load_builtins.create_listlike_dataset(mixed_string_list,h5_data,"mixed_string_list",**compression_kwargs)
    assert isinstance(h_dataset,h5.Group)
    item_name = "data{:d}"
    index = -1 
    loaded_list = load_builtins.ListLikeContainer(h_dataset.attrs,b'list',list)
    index_from_string = load_builtins.ListLikeContainer(h_dataset.attrs,b'list',list)
    for index,(name,item,attrs,kwargs) in enumerate(iter(subitems)):
        assert item_name.format(index) == name and item == mixed_string_list[index]
        assert attrs == {"item_index":index} and kwargs == compression_kwargs
        item_dataset = h_dataset.create_dataset(name,data = item)
        item_dataset.attrs.update(attrs)
        loaded_list.append(name,item,item_dataset.attrs)
        index_from_string.append(name,item,{})
    assert index + 1 == len(mixed_string_list)
    assert loaded_list.convert() == mixed_string_list
    assert index_from_string.convert() == mixed_string_list
    

def test_set_container(h5_data,compression_kwargs):
    """
    tests storing and loading of set
    """

    # check that set of strings is store as group
    test_set = {'I','confess','appriciate','hickle','times'}
    h_setdataset,subitems = load_builtins.create_setlike_dataset(test_set,h5_data,"test_set",**compression_kwargs)
    set_container = load_builtins.SetLikeContainer(h_setdataset.attrs,b'set',set)
    for name,item,attrs,kwargs in subitems:
        set_container.append(name,item,attrs)
    assert set_container.convert() == test_set

    # check that set of single bytes is stored as single dataset
    test_set_2 = set(b"hello world")
    h_setdataset,subitems = load_builtins.create_setlike_dataset(test_set_2,h5_data,"test_set_2",**compression_kwargs)
    assert isinstance(h_setdataset,h5.Dataset) and set(h_setdataset[()]) == test_set_2
    assert not subitems and iter(subitems)
    assert load_builtins.load_list_dataset(h_setdataset,b'set',set) == test_set_2

    # check that set containing byte strings is stored as group
    test_set_3 = set((item.encode("utf8") for item in test_set))
    h_setdataset,subitems = load_builtins.create_setlike_dataset(test_set_3,h5_data,"test_set_3",**compression_kwargs)
    set_container = load_builtins.SetLikeContainer(h_setdataset.attrs,b'set',set)
    for name,item,attrs,kwargs in subitems:
        set_container.append(name,item,attrs)
    assert set_container.convert() == test_set_3

    # check that empty set is represented by empty dataset
    h_setdataset,subitems = load_builtins.create_setlike_dataset(set(),h5_data,"empty_set",**compression_kwargs)
    assert isinstance(h_setdataset,h5.Dataset) and h_setdataset.size == 0
    assert not subitems and iter(subitems)
    assert load_builtins.load_list_dataset(h_setdataset,b'set',set) == set()
    

def test_dictlike_dataset(h5_data,compression_kwargs):
    """
    test storing and loading of dict
    """

    class KeyClass():
        """class used as dict key"""

    allkeys_dict = {
        'string_key':0,
        b'bytes_key':1,
        12:2,
        0.25:3,
        complex(1,2):4,
        None:5,
        (1,2,3):6,
        KeyClass():7,
        KeyClass:8
    }

    # check that dict is stored as group
    # check that string and byte string keys are mapped to dataset or group name 
    # check that scalar dict keys are converted to their string representation
    # check that for all other keys a key value pair is created
    h_datagroup,subitems = load_builtins.create_dictlike_dataset(allkeys_dict,h5_data,"allkeys_dict",**compression_kwargs)
    assert isinstance(h_datagroup,h5.Group)
    invalid_key = b''
    last_entry = -1
    load_dict = load_builtins.DictLikeContainer(h_datagroup.attrs,b'dict',dict)
    ordered_dict = collections.OrderedDict()
    for name,item,attrs,kwargs in subitems:
        value = item
        if attrs["key_base_type"] == b"str":
            key = name[1:-1]
        elif attrs["key_base_type"] == b"bytes":
            key = name[2:-1].encode("utf8")
        elif attrs["key_base_type"] == b'key_value':
            key = item[0]
            value = item[1]
        else: 
            load_key = load_builtins.dict_key_types_dict.get(attrs["key_base_type"],None)
            if load_key is None:
                raise ValueError("key_base_type '{}' invalid".format(attrs["key_base_type"]))
            key = load_key(name)
        assert allkeys_dict.get(key,invalid_key) == value
        load_dict.append(name,item,attrs)
        last_entry = attrs.get("key_idx",None)
        ordered_dict[key] = value
    assert last_entry + 1 == len(allkeys_dict) 
    assert load_dict.convert() == allkeys_dict

    # verify that DictLikeContainer.append raises error in case invalid key_base_type
    # is provided
    with pytest.raises(ValueError, match = r"key\s+type\s+'.+'\s+not\s+understood"):
        load_dict.append("invalid_key_type",12,{"key_idx":9,"key_base_type":b"invalid_type"})
    tuple_key = ('a','b','c')

    # verify that DictLikeContainer.append raises error in case index of key value pair
    # within dict is whether provided by key_index attribute nor can be parsed from 
    # name of corresponding dataset or group
    with pytest.raises(KeyError, match = r"invalid\s+dict\s+item\s+key_index\s+missing"):
        load_dict.append(str(tuple_key),9,{"item_index":9,"key_base_type":b"tuple"})

    # check that helpers.nobody_is_my_name injected for example by load_nothing is silently
    # ignored in case no key could be retrieved from dataset or sub group
    load_dict.append(
        str(tuple_key), helpers.nobody_is_my_name,
        {"item_index":9,"key_base_type":b"tuple"}
    )
    with pytest.raises(KeyError):
        assert load_dict.convert()[tuple_key] is None

    # check that if key_idx attribute is provided key value pair may be added
    load_dict.append(str(tuple_key),9,{"key_idx":9,"key_base_type":b"tuple"})
    assert load_dict.convert()[tuple_key] == 9

    # verify that DictLikeContainer.append raises error in case item index already
    # set
    with pytest.raises(IndexError,match = r"Key\s+index\s+\d+\s+already\s+set"):
        load_dict.append(str(tuple_key),9,{"key_idx":9,"key_base_type":b"tuple"})

    # check that order of OrderedDict dict keys is not altered on loading data from 
    # hickle file
    h_datagroup,subitems = load_builtins.create_dictlike_dataset(ordered_dict,h5_data,"ordered_dict",**compression_kwargs)
    assert isinstance(h_datagroup,h5.Group)
    last_entry = -1
    load_ordered_dict = load_builtins.DictLikeContainer(h_datagroup.attrs,b'dict',collections.OrderedDict)
    for name,item,attrs,kwargs in subitems:
        value = item
        if attrs["key_base_type"] == b"str":
            key = name[1:-1]
        elif attrs["key_base_type"] == b"bytes":
            key = name[2:-1].encode("utf8")
        elif attrs["key_base_type"] == b'key_value':
            key = item[0]
            value = item[1]
        else: 
            load_key = load_builtins.dict_key_types_dict.get(attrs["key_base_type"],None)
            if load_key is None:
                raise ValueError("key_base_type '{}' invalid".format(attrs["key_base_type"]))
            key = load_key(name)
        assert ordered_dict.get(key,invalid_key) == value
        load_ordered_dict.append(name,item,attrs)
        last_entry = attrs.get("key_idx",None)
    assert last_entry + 1 == len(allkeys_dict) 
    assert load_ordered_dict.convert() == ordered_dict


# %% MAIN SCRIPT
if __name__ == "__main__":
    from _pytest.fixtures import FixtureRequest
    from hickle.tests.conftest import compression_kwargs
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_scalar_dataset),)
    ):
        test_scalar_dataset(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_non_dataset),)
    ):
        test_non_dataset(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_listlike_dataset),)
    ):
        test_listlike_dataset(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_set_container),)
    ):
        test_set_container(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_dictlike_dataset),)
    ):
        test_dictlike_dataset(h5_root,keywords)

    
    

